import functools
from collections import defaultdict
from itertools import chain
from typing import Iterable
from typing import List, Dict, Optional, Tuple, ClassVar
import time
from more_itertools import flatten

from prometheus_client import Gauge

from bxcommon.connections.abstract_connection import AbstractConnection
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.utils import memory_utils
from bxcommon.utils.stats import hooks
from bxcommon.utils.stats.memory_statistics_service import memory_statistics
from bxutils import logging

logger = logging.get_logger(__name__)


class ConnectionPool:
    """
    Class for managing a set of connection handlers over a node's lifecycle.

    Connections should be thought of unique identified by via file no or node id.
    (ip, port) will sometimes be used for look ups, but can sometimes be unreliable to identify
    a connection to another node depending on who initiates the connection.

    Try to avoid writing code that needs to rely on looking up connections by (ip, port), and
    include fallbacks for node id lookup.

    Lookups without node id are still allowed to support preSSL connections that
    do not have as strong guarantees on identifying connections by node id.
    """
    INITIAL_FILENO: ClassVar[int] = 100

    by_ipport: Dict[Tuple[str, int], AbstractConnection]
    by_connection_type: Dict[ConnectionType, List[AbstractConnection]]
    by_node_id: Dict[str, AbstractConnection]
    len_fileno: int
    count_conn_by_ip: Dict[str, int]

    def __init__(self) -> None:
        self.by_fileno = [None] * ConnectionPool.INITIAL_FILENO
        self.by_ipport = {}
        self.by_connection_type = defaultdict(list)
        self.by_node_id = {}
        self.len_fileno = ConnectionPool.INITIAL_FILENO
        self.count_conn_by_ip = defaultdict(lambda: 0)

        self._create_metrics()

    def add(self, fileno: int, ip: str, port: int, conn: AbstractConnection) -> None:
        """
        Adds a connection for a tracking.
        Throws an AssertionError if there already exists a connection to the same (ip, port) pair.
        """
        assert (ip, port) not in self.by_ipport

        while fileno >= self.len_fileno:
            self.by_fileno.extend([None] * ConnectionPool.INITIAL_FILENO)
            self.len_fileno += ConnectionPool.INITIAL_FILENO

        self.by_fileno[fileno] = conn
        self.by_ipport[(ip, port)] = conn
        self.by_connection_type[conn.CONNECTION_TYPE].append(conn)
        self.count_conn_by_ip[ip] += 1

    def update_port(self, old_port: int, new_port: int, conn: AbstractConnection) -> None:
        """
        Updates port mapping of connection. Clears out old one.
        """
        old_ipport = (conn.peer_ip, old_port)
        if old_ipport in self.by_ipport:
            del self.by_ipport[old_ipport]

        self.by_ipport[(conn.peer_ip, new_port)] = conn

    def update_connection_type(self, conn: AbstractConnection, connection_type: ConnectionType) -> None:
        self.delete(conn)

        # pyre-ignore this is how we currently identify connections
        conn.CONNECTION_TYPE = connection_type
        conn.format_connection()
        self.add(conn.file_no, conn.peer_ip, conn.peer_port, conn)

        peer_id = conn.peer_id
        assert peer_id is not None
        self.index_conn_node_id(peer_id, conn)

    def index_conn_node_id(self, node_id: str, conn: AbstractConnection) -> None:
        self.by_node_id[node_id] = conn

    def has_connection(
            self, ip: Optional[str] = None, port: Optional[int] = None, node_id: Optional[str] = None) -> bool:
        if node_id is not None and node_id in self.by_node_id:
            return True
        if ip is None or port is None:
            return False
        return (ip, port) in self.by_ipport

    def get_connection_by_network_num(self, network_num: int) -> Iterable[AbstractConnection]:
        for connection in self.get_by_connection_types(
            (ConnectionType.GATEWAY, ConnectionType.RELAY_PROXY)
        ):
            if connection.network_num == network_num:
                yield connection

    def get_by_connection_types(
        self, connection_types: Tuple[ConnectionType, ...]
    ) -> List[AbstractConnection]:
        # pyre-fixme [7]: Expected `List[AbstractConnection[typing.Any]]`
        #  but got `typing.Iterator[Variable[more_itertools.recipes._T]]`.
        return flatten(self._iter_by_connection_types(connection_types))

    def _iter_by_connection_types(
        self, connection_types: Tuple[ConnectionType, ...]
    ) -> Iterable[AbstractConnection]:
        for connection_type, connections in self.by_connection_type.items():
            if any(connection_type & matching_type for matching_type in connection_types):
                # pyre-fixme[7]: Expected `Iterable[AbstractConnection[typing.Any]]`
                #  but got `typing.Generator[chain[AbstractConnection[typing.Any]], None, None]`.
                yield chain(connections)

    def get_by_ipport(self, ip: str, port: int, node_id: Optional[str] = None) -> AbstractConnection:
        ip_port = (ip, port)
        if ip_port in self.by_ipport:
            return self.by_ipport[ip_port]

        if node_id is not None and node_id in self.by_node_id:
            return self.by_node_id[node_id]

        raise KeyError(f"Could not find a connection with ip port: {ip_port} or node id: {node_id}")

    def get_by_fileno(self, fileno: int) -> Optional[AbstractConnection]:
        if fileno >= self.len_fileno:
            return None
        return self.by_fileno[fileno]

    def get_by_node_id(self, node_id: str) -> AbstractConnection:
        return self.by_node_id[node_id]

    def delete(self, conn: AbstractConnection) -> None:
        """
        Delete connection from connection pool.
        """
        # Remove conn from the dictionaries
        self.by_fileno[conn.file_no] = None

        # Connection might be replaced with new connection
        # Only delete from byipport if connection has the matching fileno
        ipport = (conn.peer_ip, conn.peer_port)
        if ipport in self.by_ipport and self.by_ipport[ipport].file_no == conn.file_no:
            del self.by_ipport[(conn.peer_ip, conn.peer_port)]

        if conn.CONNECTION_TYPE in self.by_connection_type:
            if len(self.by_connection_type[conn.CONNECTION_TYPE]) == 1:
                del self.by_connection_type[conn.CONNECTION_TYPE]
            else:
                self.by_connection_type[conn.CONNECTION_TYPE].remove(conn)

        # Decrement the count- if it's 0, we delete the key.
        if self.count_conn_by_ip[conn.peer_ip] == 1:
            del self.count_conn_by_ip[conn.peer_ip]
        else:
            self.count_conn_by_ip[conn.peer_ip] -= 1

        peer_id = conn.peer_id
        if peer_id and peer_id in self.by_node_id:
            del self.by_node_id[peer_id]

    def items(self):
        """
        Iterates through all of the connection objects in this connection pool.

        The pool can be freely modified while iterating here.
        """
        for fileno, conn in enumerate(self.by_fileno):
            if conn is not None:
                yield fileno, conn

    def __repr__(self) -> str:
        return repr(list(self.by_ipport.values()))

    def __iter__(self):
        """
        Iterates through all of the connection objects in this connection pool.

        Do not modify this pool while iterating through it here.
        """
        for ipport in self.by_ipport:
            yield self.by_ipport[ipport]

    def __len__(self):
        """
        Returns number of connections in pool.
        """
        return len(self.by_ipport)

    def log_connection_pool_mem_stats(self):
        """
        Logs Connection Pool memory statistics
        """

        sizer_obj = memory_statistics.sizer_obj
        sizer = sizer_obj.sizer
        logger.trace("MemoryStats excluded classes: {}", sizer_obj.excluded)

        by_fileno_obj_size = self._log_detailed_object_size(self.by_fileno, "by_fileno", sizer=sizer)
        by_ipport_obj_size = self._log_detailed_object_size(self.by_ipport, "by_ipport", sizer=sizer)
        by_conn_type_obj_size = self._log_detailed_object_size(self.by_connection_type, "by_conn_type", sizer=sizer)
        by_node_id_obj_size = self._log_detailed_object_size(self.by_node_id, "by_node_id", sizer=sizer)

        class_name = self.__class__.__name__
        hooks.add_obj_mem_stats(
            class_name,
            0,
            self.by_fileno,
            "connection_pool_by_fileno",
            by_fileno_obj_size,
            object_item_count=sum([1 for i in self.by_fileno if i is not None]),
            object_type=memory_utils.ObjectType.META,
            size_type=memory_utils.SizeType.OBJECT
        )

        hooks.add_obj_mem_stats(
            class_name,
            0,
            self.by_ipport,
            "connection_pool_by_ipport",
            by_ipport_obj_size,
            object_item_count=len(self.by_ipport),
            object_type=memory_utils.ObjectType.META,
            size_type=memory_utils.SizeType.OBJECT
        )

        hooks.add_obj_mem_stats(
            class_name,
            0,
            self.by_node_id,
            "connection_pool_by_node_id",
            by_node_id_obj_size,
            object_item_count=len(self.by_node_id),
            object_type=memory_utils.ObjectType.META,
            size_type=memory_utils.SizeType.OBJECT
        )

        hooks.add_obj_mem_stats(
            class_name,
            0,
            self.by_connection_type,
            "connection_pool_by_connection_type",
            by_conn_type_obj_size,
            object_item_count=len(self.by_connection_type),
            object_type=memory_utils.ObjectType.META,
            size_type=memory_utils.SizeType.OBJECT
        )

        hooks.add_obj_mem_stats(
            class_name,
            0,
            self.count_conn_by_ip,
            "connection_pool_count_conn_by_ip",
            memory_utils.get_object_size(self.count_conn_by_ip),
            object_item_count=len(self.count_conn_by_ip),
            object_type=memory_utils.ObjectType.BASE,
            size_type=memory_utils.SizeType.TRUE
        )
        self._log_connections_mem_stats()

    def _log_detailed_object_size(self, obj, stat_name, sizer):
        class_name = self.__class__.__name__
        start_time = time.time()
        obj_size = memory_utils.get_detailed_object_size(obj, sizer=sizer)
        logger.debug("(MemoryStats) ({}) {} took: {:.3f} seconds", class_name, stat_name, time.time() - start_time)
        return obj_size

    def _log_connections_mem_stats(self):
        by_fileno_size = len(self.by_fileno)
        for conn_index in range(by_fileno_size):
            conn = self.by_fileno[conn_index]
            if conn is not None:
                hooks.reset_class_mem_stats(conn.__class__.__name__)

        for conn_index in range(by_fileno_size):
            conn = self.by_fileno[conn_index]
            if conn is not None:
                conn.log_connection_mem_stats()

    # Not used right now but is helpful for debugging
    def _get_ref_info(self, parent_obj):
        return [(obj.name, obj.size, self._get_ref_info(obj)) for obj in parent_obj.references]

    def _create_metrics(self) -> None:
        self.connections_gauge = Gauge(
            "peer_count",
            "Number of peers node is connected to",
            ("connection_type",)
        )
        self.connections_gauge.labels("total").set_function(
            functools.partial(len, self.by_ipport)
        )
        self.connections_gauge.labels("blockchain").set_function(
            functools.partial(self._get_number_of_connections, ConnectionType.BLOCKCHAIN_NODE)
        )
        self.connections_gauge.labels("relay_transaction").set_function(
            functools.partial(self._get_number_of_connections, ConnectionType.RELAY_TRANSACTION)
        )
        self.connections_gauge.labels("relay_block").set_function(
            functools.partial(self._get_number_of_connections, ConnectionType.RELAY_BLOCK)
        )
        self.connections_gauge.labels("relay_all").set_function(
            functools.partial(self._get_number_of_connections, ConnectionType.RELAY_ALL)
        )

    def _get_number_of_connections(self, connection_type: ConnectionType) -> int:
        return len(list(self.get_by_connection_types((connection_type,))))



import struct
from datetime import datetime
from typing import Optional

from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxutils.logging import LogLevel


class BdnPerformanceStatsMessageV9(AbstractBloxrouteMessage):
    """
    Bloxroute message sent from gateway to relay that contains statistics on BDN performance.
    """

    MSG_SIZE = AbstractBloxrouteMessage.HEADER_LENGTH + (2 * constants.DOUBLE_SIZE_IN_BYTES) + \
               (4 * constants.UL_SHORT_SIZE_IN_BYTES) + constants.CONTROL_FLAGS_LEN
    MESSAGE_TYPE = BloxrouteMessageType.BDN_PERFORMANCE_STATS

    _interval_start_time: Optional[float] = None
    _interval_end_time: Optional[float] = None
    _new_blocks_received_from_blockchain_node: Optional[int] = None
    _new_blocks_received_from_bdn: Optional[int] = None
    _new_tx_received_from_blockchain_node: Optional[int] = None
    _new_tx_received_from_bdn: Optional[int] = None

    def __init__(
        self,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        new_blocks_received_from_blockchain_node: Optional[int] = None,
        new_blocks_received_from_bdn: Optional[int] = None,
        new_tx_received_from_blockchain_node: Optional[int] = None,
        new_tx_received_from_bdn: Optional[int] = None,
        buf: Optional[bytearray] = None
    ):
        if buf is None:
            assert start_time is not None
            assert end_time is not None
            assert new_blocks_received_from_blockchain_node is not None
            assert new_blocks_received_from_bdn is not None
            assert new_tx_received_from_blockchain_node is not None
            assert new_tx_received_from_bdn is not None

            buf = bytearray(self.MSG_SIZE)

            off = AbstractBloxrouteMessage.HEADER_LENGTH
            struct.pack_into("<d", buf, off, start_time.timestamp())
            off += constants.DOUBLE_SIZE_IN_BYTES

            struct.pack_into("<d", buf, off, end_time.timestamp())
            off += constants.DOUBLE_SIZE_IN_BYTES

            struct.pack_into("<H", buf, off, new_blocks_received_from_blockchain_node)
            off += constants.UL_SHORT_SIZE_IN_BYTES

            struct.pack_into("<H", buf, off, new_blocks_received_from_bdn)
            off += constants.UL_SHORT_SIZE_IN_BYTES

            struct.pack_into("<H", buf, off, new_tx_received_from_blockchain_node)
            off += constants.UL_SHORT_SIZE_IN_BYTES

            struct.pack_into("<H", buf, off, new_tx_received_from_bdn)
            off += constants.UL_SHORT_SIZE_IN_BYTES

        self.buf = buf
        payload_length = len(buf) - AbstractBloxrouteMessage.HEADER_LENGTH
        super().__init__(self.MESSAGE_TYPE, payload_length, self.buf)

    def log_level(self):
        return LogLevel.DEBUG

    def interval_start_time(self) -> datetime:
        if self._interval_start_time is None:
            self._unpack()

        interval_start_time = self._interval_start_time
        assert interval_start_time is not None
        return datetime.fromtimestamp(interval_start_time)

    def interval_end_time(self) -> datetime:
        if self._interval_end_time is None:
            self._unpack()

        interval_end_time = self._interval_end_time
        assert interval_end_time is not None
        return datetime.fromtimestamp(interval_end_time)

    def new_blocks_from_blockchain_node(self) -> int:
        if self._new_blocks_received_from_blockchain_node is None:
            self._unpack()

        new_blocks_received_from_blockchain_node = self._new_blocks_received_from_blockchain_node
        assert new_blocks_received_from_blockchain_node is not None
        return new_blocks_received_from_blockchain_node

    def new_blocks_from_bdn(self) -> int:
        if self._new_blocks_received_from_bdn is None:
            self._unpack()

        new_blocks_received_from_bdn = self._new_blocks_received_from_bdn
        assert new_blocks_received_from_bdn is not None
        return new_blocks_received_from_bdn

    def new_tx_from_blockchain_node(self) -> int:
        if self._new_tx_received_from_blockchain_node is None:
            self._unpack()

        new_tx_received_from_blockchain_node = self._new_tx_received_from_blockchain_node
        assert new_tx_received_from_blockchain_node is not None
        return new_tx_received_from_blockchain_node

    def new_tx_from_bdn(self) -> int:
        if self._new_tx_received_from_bdn is None:
            self._unpack()

        new_tx_received_from_bdn = self._new_tx_received_from_bdn
        assert new_tx_received_from_bdn is not None
        return new_tx_received_from_bdn

    def _unpack(self):
        off = AbstractBloxrouteMessage.HEADER_LENGTH
        self._interval_start_time, = struct.unpack_from("<d", self.buf, off)
        off += constants.DOUBLE_SIZE_IN_BYTES
        self._interval_end_time, = struct.unpack_from("<d", self.buf, off)
        off += constants.DOUBLE_SIZE_IN_BYTES
        self._new_blocks_received_from_blockchain_node, = struct.unpack_from("<H", self.buf, off)
        off += constants.UL_SHORT_SIZE_IN_BYTES
        self._new_blocks_received_from_bdn, = struct.unpack_from("<H", self.buf, off)
        off += constants.UL_SHORT_SIZE_IN_BYTES
        self._new_tx_received_from_blockchain_node, = struct.unpack_from("<H", self.buf, off)
        off += constants.UL_SHORT_SIZE_IN_BYTES
        self._new_tx_received_from_bdn, = struct.unpack_from("<H", self.buf, off)
        off += constants.UL_SHORT_SIZE_IN_BYTES

    def __repr__(self):
        return "BdnPerformanceStatsMessage<blocks_from_blockchain_node: {}, blocks_from_bdn: {}, " \
               "tx_from_blockchain_node: {}, tx_from_bdn: {}>". \
            format(self.new_blocks_from_blockchain_node(), self.new_blocks_from_bdn(),
                   self.new_tx_from_blockchain_node(), self.new_tx_from_bdn())



import datetime
from typing import Optional, List, TYPE_CHECKING

from bxutils import logging
from bxutils.logging.log_level import LogLevel

from bxcommon.utils.stats.stat_event import StatEvent
from bxcommon.utils.stats.stat_event_type_settings import StatEventTypeSettings

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from bxcommon.connections.abstract_connection import AbstractConnection


# TODO: change default log level from STATS to info


class StatisticsEventService:
    def __init__(self) -> None:
        self.name = None
        self.log_level = LogLevel.STATS
        self.logger = logging.get_logger(__name__)
        self.priority_logger = logging.get_logger(__name__)
        self.node = None
        self.node_id: Optional[str] = None

    def set_node(self, node) -> None:
        self.node = node
        assert node.opts is not None
        self.node_id = node.opts.node_id

    def log_event(
        self,
        event_settings: StatEventTypeSettings,
        object_id: str,
        start_date_time: Optional[datetime.datetime],
        end_date_time: Optional[datetime.datetime],
        peers: Optional[List["AbstractConnection"]] = None,
        **kwargs
    ) -> None:
        node_id = self.node_id
        assert node_id is not None

        if start_date_time is None:
            start_date_time = datetime.datetime.utcnow()

        if end_date_time is None:
            end_date_time = start_date_time

        stat_event = StatEvent(
            event_settings, object_id, node_id, start_date_time, end_date_time, peers, **kwargs
        )
        if event_settings.priority:
            self.priority_logger.log(self.log_level, {"data": stat_event, "type": self.name})
        else:
            self.logger.log(self.log_level, {"data": stat_event, "type": self.name})



import sys

P2P_PROTOCOL_VERSION = 4
ETH_PROTOCOL_VERSION = 65
SUPPORTED_ETH_PROTOCOL_VERSION = [ETH_PROTOCOL_VERSION, 64, 63]
CAPABILITIES = ((b"eth", ETH_PROTOCOL_VERSION),)
BX_ETH_CLIENT_NAME = b"BloxRoute Gateway"

SIGNATURE_LEN = 65
MDC_LEN = 32
MAC_LEN = 32

MSG_TYPE_LEN = 1
MSG_HEADER_OFF = 0

PRIVATE_KEY_LEN = 32
PUBLIC_KEY_LEN = 64
PUBLIC_KEY_X_Y_LEN = 32

ECIES_CIPHER_NAME = "aes-128-ctr"
RLPX_CIPHER_NAME = "aes-256-ctr"
ECIES_CURVE = "secp256k1"
ECIES_ENCRYPT_OVERHEAD_LENGTH = 113
ECIES_HEADER = 0x04
ECIES_HEADER_BYTE = b"\x04"
ECIES_HEADER_LEN = 1

KEY_MATERIAL_LEN = 32
SHARED_KEY_LEN = 32
ENC_KEY_LEN = 16
IV_LEN = 16

CIPHER_DECRYPT_DO = 0
CIPHER_ENCRYPT_DO = 1

PING_MSG_TTL_SEC = 60

FRAME_HDR_TOTAL_LEN = 32
FRAME_HDR_DATA_LEN = 16
FRAME_MAC_LEN = 16
MSG_PADDING = 16
FRAME_MAX_BODY_SIZE = 256 ** 3
FRAME_MSG_TYPE_LEN = 1

DEFAULT_FRAME_SIZE = sys.maxsize
DEFAULT_FRAME_PROTOCOL_ID = 0
MAX_FRAME_SEQUENCE_ID = 2 ** 16

ENC_AUTH_MSG_LEN = 307
EIP8_AUTH_PREFIX_LEN = 2
EIP8_ACK_PAD_MIN = 100
EIP8_ACK_PAD_MAX = 250
AUTH_MSG_LEN = 194
AUTH_ACK_MSG_LEN = 97
ENC_AUTH_ACK_MSG_LEN = 210
AUTH_NONCE_LEN = 32
AUTH_MSG_VERSION = 4
MAX_NONCE = 2 ** 256 - 1

SHA3_LEN_BYTES = 32
SHA3_LEN_BITS = 256

BLOCK_HASH_LEN = 32
ADDRESS_LEN = 20
MERKLE_ROOT_LEN = 32
BLOOM_LEN = 256
BLOCK_NONCE_LEN = 8
FIXED_LENGTH_FIELD_OFFSET = 2 * BLOCK_HASH_LEN + ADDRESS_LEN + 3 * MERKLE_ROOT_LEN + BLOOM_LEN + 9

HANDSHAKE_TIMEOUT_SEC = 30
PING_PONG_INTERVAL_SEC = 30
PING_PONG_TIMEOUT_SEC = 60
DISCONNECT_DELAY_SEC = 2
DISCOVERY_PONG_TIMEOUT_SEC = 5

DISCONNECT_REASON_TIMEOUT = 0x0b

CHECKPOINT_BLOCK_HEADERS_REQUEST_WAIT_TIME_S = 5

MSG_CLS_SERIALIZER_ATTR = "serializer"

BLOCK_CLEANUP_NODE_BLOCK_LIST_POLL_INTERVAL_S = 0
TRACKED_BLOCK_CLEANUP_INTERVAL_S = 15

LAST_KNOWN_TOTAL_DIFFICULTIES_MAX_COUNT = 10
NEW_BLOCK_PARTS_MAX_WAIT_S = 5
REQUESTED_NEW_BLOCK_HEADERS_MAX_COUNT = 10
REQUESTED_NEW_BLOCK_BODIES_MAX_COUNT = 10

CHECK_BLOCK_RECEIPT_DELAY_S = 0.1
CHECK_BLOCK_RECEIPT_INTERVAL_S = 0.1
CHECK_BLOCK_RECEIPT_MAX_COUNT = 5

GATEWAY_PRIVATE_KEY_FILE_NAME = ".gatewaykey"
BLOCK_CONFIRMATION_REQUEST_INTERVALS = 5
BLOCK_CONFIRMATION_REQUEST_CACHE_INTERVAL_S = 30 * 60

EIP155_CHAIN_ID_OFFSET = 35

ETH_DEFAULT_BLOCK_SIZE = 100000
ETH_MINIMAL_SUB_TASK_TX_COUNT = 100

# skip transaction messages for 30s if they're greater than 30k in size
# Geth packs sync messages to 100k, should check other implementations
# these sync messages are usually useless to the BDN
ETH_SKIP_TRANSACTIONS_LEAD_TIME_S = 30
ETH_SKIP_TRANSACTIONS_SIZE = 30000

MAX_FUTURE_BLOCK_NUMBER: int = 100
MAX_DIFFICULTY_CHANGE_PERCENT: int = 10

MAX_TRANSACTION_TYPE = 0xBF



from dataclasses import dataclass
from typing import Any, Dict

from bxcommon import constants
from bxcommon.models.bdn_account_model_base import Tiers
from bxcommon.models.blockchain_network_environment import BlockchainNetworkEnvironment
from bxcommon.models.blockchain_network_type import BlockchainNetworkType


@dataclass
class BlockchainNetworkModel:
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    protocol: str = None
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    network: str = None
    network_num: int = constants.UNASSIGNED_NETWORK_NUMBER
    # pyre-fixme[8]: Attribute has type `BlockchainNetworkType`; used as `None`.
    type: BlockchainNetworkType = None
    # pyre-fixme[8]: Attribute has type `BlockchainNetworkEnvironment`; used as `None`.
    environment: BlockchainNetworkEnvironment = None
    # pyre-fixme[8]: Attribute has type `Dict[str, typing.Any]`; used as `None`.
    default_attributes: Dict[str, Any] = None
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    block_interval: int = None
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    ignore_block_interval_count: int = None
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    block_recovery_timeout_s: int = None
    block_hold_timeout_s: float = constants.DEFAULT_BLOCK_HOLD_TIMEOUT
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    final_tx_confirmations_count: int = None
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    tx_contents_memory_limit_bytes: int = None
    max_block_size_bytes: int = constants.DEFAULT_MAX_PAYLOAD_LEN_BYTES
    max_tx_size_bytes: int = constants.DEFAULT_MAX_PAYLOAD_LEN_BYTES
    block_confirmations_count: int = constants.BLOCK_CONFIRMATIONS_COUNT
    tx_percent_to_log_by_hash: float = constants.TRANSACTIONS_BY_HASH_PERCENTAGE_TO_LOG_STATS_FOR
    tx_percent_to_log_by_sid: float = constants.TRANSACTIONS_BY_SID_PERCENTAGE_TO_LOG_STATS_FOR
    removed_transactions_history_expiration_s: int = constants.REMOVED_TRANSACTIONS_HISTORY_EXPIRATION_S
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    sdn_id: str = None
    tx_sync_interval_s: float = constants.GATEWAY_SYNC_TX_THRESHOLD_S
    tx_sync_sync_content: bool = constants.GATEWAY_SYNC_SYNC_CONTENT
    enable_network_content_logs: bool = False
    enable_block_compression: bool = True
    # Denoted in GWEI for Ethereum
    min_tx_network_fee: int = 0
    medium_tx_network_fee: int = 0
    min_tx_age_seconds: float = 0.0

    mempool_expected_transactions_count: int = 0
    send_compressed_txs_after_block: bool = False
    log_compressed_block_debug_info_on_relay: bool = True

    enable_recording_tx_detection_time_location: bool = True
    enable_check_sender_nonce: bool = True
    allowed_time_reuse_sender_nonce: float = 0.0
    allowed_gas_price_change_reuse_sender_nonce: float = 1.1
    enable_tx_trace: bool = False
    inject_poa: bool = False
    allowed_from_tier: str = Tiers.DEVELOPER.value



import blxr_rlp as rlp
from bxgateway.messages.eth.discovery.eth_discovery_message import EthDiscoveryMessage
from bxgateway.messages.eth.discovery.eth_discovery_message_type import EthDiscoveryMessageType
from bxgateway.messages.eth.serializers.endpoint import Endpoint


class PongEthDiscoveryMessage(EthDiscoveryMessage):
    msg_type = EthDiscoveryMessageType.PONG

    fields = [("remote_address", Endpoint()),
              ("ping_hash", rlp.sedes.raw),
              ("expiration", rlp.sedes.big_endian_int)]

    def get_remote_address(self):
        return self.get_field_value("remote_address")

    def get_ping_hash(self):
        return self.get_field_value("ping_hash")

    def get_expiration(self):
        return self.get_field_value("expiration")



import time
from typing import List, TYPE_CHECKING

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.utils import convert
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import NULL_BTC_BLOCK_HASH
from bxcommon.utils.expiring_dict import ExpiringDict
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats.block_stat_event_type import BlockStatEventType
from bxcommon.utils.stats.block_statistics_service import block_stats
from bxgateway import gateway_constants, btc_constants
from bxgateway.btc_constants import NODE_WITNESS_SERVICE_FLAG
from bxgateway.connections.btc.btc_base_connection_protocol import BtcBaseConnectionProtocol
from bxgateway import log_messages
from bxgateway.messages.btc.block_transactions_btc_message import BlockTransactionsBtcMessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType
from bxgateway.messages.btc.compact_block_btc_message import CompactBlockBtcMessage
from bxgateway.messages.btc.data_btc_message import GetBlocksBtcMessage
from bxgateway.messages.btc.get_block_transactions_btc_message import GetBlockTransactionsBtcMessage
from bxgateway.messages.btc.inventory_btc_message import GetDataBtcMessage, InventoryType, InvBtcMessage
from bxgateway.messages.btc.send_compact_btc_message import SendCompactBtcMessage
from bxgateway.messages.btc.ver_ack_btc_message import VerAckBtcMessage
from bxgateway.messages.btc.version_btc_message import VersionBtcMessage
from bxgateway.utils.errors.message_conversion_error import MessageConversionError

if TYPE_CHECKING:
    from bxgateway.connections.btc.btc_node_connection import BtcNodeConnection


class BtcNodeConnectionProtocol(BtcBaseConnectionProtocol):
    def __init__(self, connection: "BtcNodeConnection"):
        super(BtcNodeConnectionProtocol, self).__init__(connection)

        connection.message_handlers.update({
            BtcMessageType.VERSION: self.msg_version,
            BtcMessageType.INVENTORY: self.msg_inv,
            BtcMessageType.BLOCK: self.msg_block,
            BtcMessageType.TRANSACTIONS: self.msg_tx,
            BtcMessageType.GET_BLOCKS: self.msg_get_blocks,
            BtcMessageType.GET_HEADERS: self.msg_get_headers,
            BtcMessageType.GET_DATA: self.msg_get_data,
            BtcMessageType.REJECT: self.msg_reject,
            BtcMessageType.COMPACT_BLOCK: self.msg_compact_block,
            BtcMessageType.BLOCK_TRANSACTIONS: self.msg_block_transactions
        })

        self.request_witness_data = False
        self._recovery_compact_blocks = ExpiringDict(
            self.node.alarm_queue,
            btc_constants.BTC_COMPACT_BLOCK_RECOVERY_TIMEOUT_S,
            f"{str(self)}_compact_btc_recoveries"
        )
        self.ping_interval_s: int = gateway_constants.BLOCKCHAIN_PING_INTERVAL_S
        self.connection.node.alarm_queue.register_alarm(
            self.block_cleanup_poll_interval_s,
            self._request_blocks_confirmation
        )

    def msg_version(self, msg: VersionBtcMessage) -> None:
        """
        Handle version message.
        Gateway initiates connection, so do not check for misbehavior. Record that we received the version message,
        send a verack, and synchronize chains if need be.
        :param msg: VERSION message
        """
        self.request_witness_data = msg.services() & NODE_WITNESS_SERVICE_FLAG > 0

        if self.request_witness_data:
            self.connection.log_debug("Connection with Bitcoin node supports SegWit.")

        self.connection.on_connection_established()
        reply = VerAckBtcMessage(self.magic)
        self.connection.enqueue_msg(reply)

        send_compact_msg = SendCompactBtcMessage(
            self.magic, on_flag=self.node.opts.compact_block, version=1
        )

        self.node.alarm_queue.register_alarm(
            2, self.connection.enqueue_msg, send_compact_msg
        )
        self.connection.schedule_pings()

        if self.connection.is_active():
            self.node.on_blockchain_connection_ready(self.connection)

    def msg_inv(self, msg: InvBtcMessage) -> None:
        """
        Handle an inventory message.

        Requests all transactions and blocks that haven't been previously seen.
        :param msg: INV message
        """
        contains_block = False
        inventory_requests = []
        block_hashes = []
        for inventory_type, item_hash in msg:
            if InventoryType.is_block(inventory_type):
                if not self.node.should_process_block_hash(item_hash):
                    continue
                block_hashes.append(item_hash)
                if item_hash not in self.node.blocks_seen.contents:
                    contains_block = True
                    inventory_requests.append((inventory_type, item_hash))
            else:
                inventory_requests.append((inventory_type, item_hash))

        self.node.block_cleanup_service.mark_blocks_and_request_cleanup(block_hashes)

        if inventory_requests:
            get_data = GetDataBtcMessage(
                magic=msg.magic(),
                inv_vects=inventory_requests,
                request_witness_data=self.request_witness_data
            )
            self.connection.enqueue_msg(get_data, prepend=contains_block)

        block_queuing_service = self.node.block_queuing_service_manager.get_block_queuing_service(self.connection)
        if block_queuing_service is not None:
            block_queuing_service.mark_blocks_seen_by_blockchain_node(block_hashes)

    def msg_get_data(self, msg: GetDataBtcMessage) -> None:
        """
        Handle GETDATA message from Bitcoin node.
        :param msg: GETDATA message
        """

        for inv_type, object_hash in msg:
            if InventoryType.is_block(inv_type):
                block_stats.add_block_event_by_block_hash(
                    object_hash,
                    BlockStatEventType.REMOTE_BLOCK_REQUESTED_BY_GATEWAY,
                    network_num=self.connection.network_num,
                    more_info="Protocol: {}, Network: {}".format(
                        self.node.opts.blockchain_protocol,
                        self.node.opts.blockchain_network
                    )
                )
            inv_msg = InvBtcMessage(
                magic=self.magic, inv_vects=[(InventoryType.MSG_BLOCK, object_hash)]
            )
            self.connection.enqueue_msg(inv_msg)
        return self.msg_proxy_request(msg, self.connection)

    def msg_get_blocks(self, msg):
        self.msg_proxy_request(msg, self.connection)

    def msg_get_headers(self, msg):
        self.msg_proxy_request(msg, self.connection)

    def msg_reject(self, msg):
        """
        Handle REJECT message from Bitcoin node
        :param msg: REJECT message
        """

        # Send inv message to the send in case of rejected block
        # remaining sync communication will proxy to remote blockchain node
        if msg.message() == BtcMessageType.BLOCK:
            inv_msg = InvBtcMessage(
                magic=self.magic, inv_vects=[(InventoryType.MSG_BLOCK, msg.obj_hash())]
            )
            self.connection.enqueue_msg(inv_msg)

    def msg_compact_block(self, msg: CompactBlockBtcMessage) -> None:
        """
        Handle COMPACT BLOCK message from Bitcoin node
        :param msg: COMPACT BLOCK message
        """

        block_hash = msg.block_hash()
        if not self.node.should_process_block_hash(block_hash):
            return

        short_ids_count = len(msg.short_ids())
        block_stats.add_block_event_by_block_hash(
            block_hash,
            BlockStatEventType.COMPACT_BLOCK_RECEIVED_FROM_BLOCKCHAIN_NODE,
            network_num=self.connection.network_num,
            peer=self.connection.peer_desc,
            more_info="{} short ids".format(short_ids_count)
        )

        if block_hash in self.node.blocks_seen.contents:
            self.node.on_block_seen_by_blockchain_node(block_hash)
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.COMPACT_BLOCK_RECEIVED_FROM_BLOCKCHAIN_NODE_IGNORE_SEEN,
                network_num=self.connection.network_num,
                peer=self.connection.peer_desc
            )
            self.connection.log_info(
                "Discarding duplicate block {} from local blockchain node.",
                block_hash
            )
            return

        max_time_offset = self.node.opts.blockchain_block_interval * self.node.opts.blockchain_ignore_block_interval_count
        if time.time() - msg.timestamp() >= max_time_offset:
            self.connection.log_trace(
                "Received block {} more than {} seconds after it was created ({}). Ignoring.",
                block_hash, max_time_offset, msg.timestamp()
            )
            return

        self.node.track_block_from_node_handling_started(block_hash)

        if short_ids_count < self.node.opts.compact_block_min_tx_count:
            self.connection.log_debug(
                "Compact block {} contains {} short transactions, less than limit {}. Requesting full block.",
                convert.bytes_to_hex(msg.block_hash().binary),
                short_ids_count,
                btc_constants.BTC_COMPACT_BLOCK_DECOMPRESS_MIN_TX_COUNT
            )
            get_data_msg = GetDataBtcMessage(
                magic=self.magic,
                inv_vects=[(InventoryType.MSG_BLOCK, msg.block_hash())]
            )
            self.connection.enqueue_msg(get_data_msg)
            block_stats.add_block_event_by_block_hash(block_hash,
                                                      BlockStatEventType.COMPACT_BLOCK_REQUEST_FULL,
                                                      network_num=self.connection.network_num)
            return

        self.node.block_cleanup_service.on_new_block_received(msg.block_hash(), msg.prev_block_hash())
        self.node.blocks_seen.add(block_hash)
        block_queuing_service = self.node.block_queuing_service_manager.get_block_queuing_service(self.connection)
        if block_queuing_service is not None:
            block_queuing_service.mark_block_seen_by_blockchain_node(
                block_hash,
                None,
                None
            )
        self.connection.log_info(
            "Processing compact block {} from local Bitcoin node.",
            block_hash
        )

        try:
            parse_result = self.node.block_processing_service.process_compact_block(msg, self.connection)
        except MessageConversionError as e:
            block_stats.add_block_event_by_block_hash(
                e.msg_hash,
                BlockStatEventType.BLOCK_CONVERSION_FAILED,
                network_num=self.connection.network_num,
                conversion_type=e.conversion_type.value
            )
            self.connection.log_warning(log_messages.PROCESS_BLOCK_FAILURE,
                                        e.msg_hash, e)
            get_data_msg = GetDataBtcMessage(
                magic=self.magic,
                inv_vects=[(InventoryType.MSG_BLOCK, msg.block_hash())]
            )
            self.connection.enqueue_msg(get_data_msg)
            return

        if not parse_result.success:
            self._recovery_compact_blocks.add(block_hash, parse_result)

            get_block_txs_msg = GetBlockTransactionsBtcMessage(magic=self.magic, block_hash=block_hash,
                                                               indices=parse_result.missing_indices)
            self.connection.enqueue_msg(get_block_txs_msg)

    def msg_block_transactions(self, msg: BlockTransactionsBtcMessage) -> None:
        """
        Handle BLOCK TRANSACTIONS message from Bitcoin node.
        This is the message that is sent in reply to GET BLOCK TRANSACTIONS message.
        This message exchange happens if gateway is unable to parse compact block from Bitcoin node.
        :param msg: BLOCK TRANSACTIONS message
        """

        if msg.block_hash() in self._recovery_compact_blocks.contents:
            recovery_result = self._recovery_compact_blocks.contents[msg.block_hash()]
            self.node.block_processing_service.process_compact_block_recovery(
                msg, recovery_result, self.connection
            )

    def _build_get_blocks_message_for_block_confirmation(self, hashes: List[Sha256Hash]) -> AbstractMessage:
        return GetBlocksBtcMessage(
            version=self.version,
            magic=self.magic,
            hashes=hashes,
            hash_stop=NULL_BTC_BLOCK_HASH
        )



from typing import NamedTuple

from bxcommon import constants
from bxcommon.utils import convert
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.feed.feed import Feed


class RawBlockFeedEntry:
    hash: str
    block: str

    # pylint: disable=redefined-builtin
    def __init__(
        self,
        hash: Sha256Hash,
        block: memoryview
    ) -> None:
        self.hash = str(hash)
        self.block = convert.bytes_to_hex(block)

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, RawBlockFeedEntry)
            and other.hash == self.hash
            and other.block == self.block
        )


class RawBlock(NamedTuple):
    hash: Sha256Hash
    block: memoryview


class NewBlockFeed(Feed[RawBlockFeedEntry, RawBlock]):
    NAME = "newBlocks"
    FIELDS = ["hash", "block"]
    ALL_FIELDS = FIELDS

    def __init__(self, network_num: int = constants.ALL_NETWORK_NUM,) -> None:
        super().__init__(self.NAME, network_num=network_num)

    def serialize(self, raw_message: RawBlock) -> RawBlockFeedEntry:
        return RawBlockFeedEntry(raw_message.hash, raw_message.block)



from bxcommon.messages.abstract_internal_message import AbstractInternalMessage
from bxcommon.messages.versioning.abstract_message_converter import AbstractMessageConverter


class _NoChangesMessageConverter(AbstractMessageConverter):

    def convert_to_older_version(self, msg: AbstractInternalMessage) -> AbstractInternalMessage:
        return msg

    def convert_from_older_version(self, msg: AbstractInternalMessage) -> AbstractInternalMessage:
        return msg

    def convert_first_bytes_to_older_version(self, first_msg_bytes):
        return first_msg_bytes

    def convert_first_bytes_from_older_version(self, first_msg_bytes):
        return first_msg_bytes

    def convert_last_bytes_to_older_version(self, last_msg_bytes):
        return last_msg_bytes

    def convert_last_bytes_from_older_version(self, last_msg_bytes):
        return last_msg_bytes

    def get_message_size_change_to_older_version(self):
        return 0

    def get_message_size_change_from_older_version(self):
        return 0


no_changes_message_converter = _NoChangesMessageConverter()



import asyncio
import gc
import sys
import os

from typing import Iterable, Optional, Type, Callable, List

import uvloop

from bxcommon.connections.abstract_node import AbstractNode
from bxcommon.exceptions import TerminationError, HighMemoryError, FeedSubscriptionTimeoutError
from bxcommon.models.node_type import NodeType
from bxcommon.network.node_event_loop import NodeEventLoop
from bxcommon.services import sdn_http_service
from bxcommon.utils import config
from bxcommon.common_opts import CommonOpts
from bxcommon import common_init_tasks
from bxutils import log_messages
from bxutils import logging
from bxutils.logging import log_config, LoggerConfig, gc_logger
from bxutils.logging.log_level import LogLevel
from bxutils.services.node_ssl_service import NodeSSLService
from bxutils.ssl.data import ssl_data_factory
from bxutils.ssl.ssl_certificate_type import SSLCertificateType

InitTaskType = Callable[[CommonOpts, NodeSSLService], None]
OptsType = CommonOpts
logger = logging.get_logger(__name__)

LOGGER_NAMES = ["bxcommon", "bxutils"]
THIRD_PARTY_LOGGERS = [
    LoggerConfig("urllib3", "%", LogLevel.WARNING),
    LoggerConfig("asyncio", "%", LogLevel.ERROR),
]

asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())  # pyre-ignore
asyncio.set_event_loop(uvloop.new_event_loop())


def default_ssl_service_factory(
    node_type: NodeType, ca_cert_url: str, private_ssl_base_url: str, data_dir: str
) -> NodeSSLService:
    storage_info = ssl_data_factory.create_storage_info(
        node_type=node_type,
        ca_cert_url=ca_cert_url,
        private_ssl_base_url=private_ssl_base_url,
        data_dir=data_dir,
    )
    return NodeSSLService(node_type, storage_info)


def run_node(
    process_id_file_path: str,
    opts: OptsType,
    get_node_class: Callable[[], Type[AbstractNode]],
    node_type: NodeType,
    logger_names: Optional[Iterable[str]] = tuple(LOGGER_NAMES),
    ssl_service_factory: Callable[
        [NodeType, str, str, str], NodeSSLService
    ] = default_ssl_service_factory,
    third_party_loggers: Optional[List[LoggerConfig]] = None,
    node_init_tasks: Optional[List[InitTaskType]] = None,
) -> None:

    if third_party_loggers is None:
        third_party_loggers = THIRD_PARTY_LOGGERS
    opts.logger_names = logger_names
    log_config.setup_logging(
        opts.log_format,
        opts.log_level,
        # pyre-fixme[6]: Expected `Iterable[str]` for 3rd param but got
        #  `Iterable[Optional[str]]`.
        logger_names,
        opts.log_level_overrides,
        enable_fluent_logger=opts.log_fluentd_enable,
        fluentd_host=opts.log_fluentd_host,
        fluentd_queue_size=opts.log_fluentd_queue_size,
        third_party_loggers=third_party_loggers,
        fluent_log_level=opts.log_level_fluentd,
        stdout_log_level=opts.log_level_stdout,
        fluentd_tag_suffix=node_type.name.lower()
    )
    if node_init_tasks is None:
        node_init_tasks = common_init_tasks.init_tasks

    startup_param = sys.argv[1:]
    logger.info("Startup Parameters are: {}", " ".join(startup_param))

    _verify_environment()

    config.log_pid(process_id_file_path)
    gc.callbacks.append(gc_logger.gc_callback)
    # we disable GC generation cleanup.
    # use gc.collect() if memory exceeds threshold
    gc.disable()
    try:
        if opts.use_extensions:
            from bxcommon.utils.proxy import task_pool_proxy

            task_pool_proxy.init(opts.thread_pool_parallelism_degree)
            logger.debug(
                "Initialized task thread pool parallelism degree to {}.",
                task_pool_proxy.get_pool_size(),
            )

        _run_node(
            opts,
            get_node_class,
            node_type,
            node_init_tasks=node_init_tasks,
            ssl_service_factory=ssl_service_factory,
        )
    except TerminationError:
        logger.fatal("Node terminated")
    except HighMemoryError:
        logger.info("Restarting node due to high memory")
        _close_handles()
        python = sys.executable
        os.execl(python, python, *sys.argv)
    except FeedSubscriptionTimeoutError:
        logger.info(
            "Restarting node due to feed subscription timeout, "
            "node is probably not synced with the blockchain node."
        )
        _close_handles()
        python = sys.executable
        os.execl(python, python, *sys.argv)
    except Exception as e:  # pylint: disable=broad-except
        logger.fatal("Unhandled exception {} raised, terminating!", e)

    _close_handles()


def _close_handles() -> None:
    for handler in logger.handlers:
        if hasattr(handler, "close"):
            handler.close()


def _run_node(
    opts: OptsType,
    get_node_class,
    node_type,
    node_init_tasks: List[InitTaskType],
    ssl_service_factory: Callable[
        [NodeType, str, str, str], NodeSSLService
    ] = default_ssl_service_factory,
) -> None:

    opts.node_type = node_type
    node_ssl_service = _init_ssl_service(
        node_type,
        opts.ca_cert_url,
        opts.private_ssl_base_url,
        opts.data_dir,
        ssl_service_factory=ssl_service_factory,
    )

    for task in node_init_tasks:
        task(opts, node_ssl_service)

    if not hasattr(opts, "outbound_peers"):
        opts.__dict__["outbound_peers"] = []

    logger.debug({"type": "node_init", "data": opts})

    # Start main loop
    node = get_node_class()(opts, node_ssl_service)
    log_config.set_instance(node.opts.node_id)
    loop = asyncio.get_event_loop()
    node_event_loop = NodeEventLoop(node)

    logger.trace("Running node...")
    loop.run_until_complete(node_event_loop.run())


def _init_ssl_service(
    node_type: NodeType,
    ca_cert_url: str,
    private_ssl_base_url: str,
    data_dir: str,
    ssl_service_factory: Callable[
        [NodeType, str, str, str], NodeSSLService
    ] = default_ssl_service_factory,
) -> NodeSSLService:
    node_ssl_service = ssl_service_factory(
        node_type, ca_cert_url, private_ssl_base_url, data_dir
    )
    node_ssl_service.blocking_load()

    if node_ssl_service.has_valid_certificate(SSLCertificateType.PRIVATE) \
        and not node_ssl_service.should_renew_node_certificate():
        ssl_context = node_ssl_service.create_ssl_context(SSLCertificateType.PRIVATE)
    else:
        ssl_context = node_ssl_service.create_ssl_context(
            SSLCertificateType.REGISTRATION_ONLY
        )

    sdn_http_service.reset_pool(ssl_context)
    return node_ssl_service


def _verify_environment() -> None:
    if sys.version.startswith("3.6."):
        logger.warning(log_messages.DETECTED_PYTHON3_6)



from cryptography import utils
from cryptography.x509.extensions import UnrecognizedExtension

from bxutils.ssl.extensions.extensions_object_ids import ExtensionsObjectIds


class NodePrivilegesExtension(UnrecognizedExtension):

    def __init__(self, node_privileges: str) -> None:
        super().__init__(ExtensionsObjectIds.NODE_PRIVILEGES, node_privileges.encode("utf-8"))
        self._node_privileges = node_privileges

    node_privileges = utils.read_only_property("_node_privileges")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} <node_privileges: {self._node_privileges}>"

    def __eq__(self, other) -> bool:
        if not isinstance(other, NodePrivilegesExtension):
            return NotImplemented

        return self._node_privileges == other._node_privileges

    def __ne__(self, other) -> bool:
        return not self == other

    def __hash__(self):
        return hash(self._node_privileges)



from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.v10.bdn_performance_stats_message_converter_v10 \
    import bdn_performance_stats_message_converter_v10
from bxcommon.messages.bloxroute.v11.message_converter_factory_v11 import message_converter_factory_v11
from bxcommon.messages.versioning.abstract_version_converter_factory import AbstractMessageConverterFactory
from bxcommon.messages.versioning.no_changes_message_converter import no_changes_message_converter


class _MessageConverterFactoryV10(AbstractMessageConverterFactory):
    _MESSAGE_CONVERTER_MAPPING = {
        # pylint: disable=protected-access
        **message_converter_factory_v11._MESSAGE_CONVERTER_MAPPING,
        BloxrouteMessageType.BDN_PERFORMANCE_STATS: bdn_performance_stats_message_converter_v10,
    }

    def get_message_converter(self, msg_type):
        if not msg_type:
            raise ValueError("msg_type is required.")

        if msg_type not in self._MESSAGE_CONVERTER_MAPPING:
            return no_changes_message_converter

        return self._MESSAGE_CONVERTER_MAPPING[msg_type]


message_converter_factory_v10 = _MessageConverterFactoryV10()



from typing import List, Union, Optional, cast, TYPE_CHECKING, Iterator

from bxcommon.exceptions import ChecksumError
from bxcommon.models.broadcast_message_type import BroadcastMessageType
from bxcommon.utils.blockchain_utils.ont.ont_object_hash import OntObjectHash
from bxcommon.utils.expiring_dict import ExpiringDict
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats import stats_format
from bxcommon.utils.stats.block_stat_event_type import BlockStatEventType
from bxcommon.utils.stats.block_statistics_service import block_stats
from bxgateway import ont_constants, gateway_constants
from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection
from bxgateway.messages.ont.block_ont_message import BlockOntMessage
from bxgateway.messages.ont.consensus_ont_message import OntConsensusMessage
from bxgateway.messages.ont.headers_ont_message import HeadersOntMessage
from bxgateway.messages.ont.inventory_ont_message import InvOntMessage
from bxgateway.messages.ont.inventory_ont_message import InventoryOntType
from bxgateway.services.abstract_block_queuing_service import (
    AbstractBlockQueuingService,
)
from bxutils import logging

if TYPE_CHECKING:
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode
    from bxgateway.connections.ont.ont_gateway_node import OntGatewayNode

logger = logging.get_logger(__name__)


class OntBlockQueuingService(
    AbstractBlockQueuingService[Union[BlockOntMessage, OntConsensusMessage], HeadersOntMessage]
):
    """
    Blocks sent to blockchain node only upon request
    """

    _block_hashes_by_height: ExpiringDict[int, Sha256Hash]
    _highest_block_number: int = 0

    def __init__(
        self,
        node: "AbstractGatewayNode",
        connection: AbstractGatewayBlockchainConnection
    ):
        super().__init__(node, connection)
        self.node: "OntGatewayNode" = cast("OntGatewayNode", node)
        self._block_hashes_by_height = ExpiringDict(
            node.alarm_queue,
            gateway_constants.MAX_BLOCK_CACHE_TIME_S,
            "ont_block_queue_hashes_by_heights",
        )

    def build_block_header_message(
        self, block_hash: Sha256Hash, block_message: BlockOntMessage
    ) -> HeadersOntMessage:
        return HeadersOntMessage(
            magic=block_message.magic(), headers=[block_message.header()]
        )

    # pyre-fixme[14]: `bxgateway.services.ont.ont_block_queuing_service.OntBlockQueuingService.push`
    #  overrides method defined in `AbstractBlockQueuingService` inconsistently.
    #  Parameter of type `Union[BlockOntMessage, OntConsensusMessage]` is not a supertype of the
    #  overridden parameter `Optional[Variable[bxgateway.services.abstract_block_queuing_service.TBlockMessage
    #  (bound to bxcommon.messages.abstract_block_message.AbstractBlockMessage)]]`.
    def push(
        self,
        block_hash: Sha256Hash,
        # pyre-fixme[9]: block_msg is declared to have type `Union[BlockOntMessage, OntConsensusMessage]`
        #  but is used as type `None`.
        block_msg: Union[BlockOntMessage, OntConsensusMessage] = None,
        waiting_for_recovery: bool = False,
    ):
        if self.node.opts.is_consensus and isinstance(block_msg, BlockOntMessage):
            return
        super().push(block_hash, block_msg, waiting_for_recovery)
        self.connection.log_debug(
            "Added block {} to queuing service (waiting for recovery: {})", block_hash, waiting_for_recovery
        )
        self._clean_block_queue()

        if block_msg is None:
            return

        try:
            block_msg.validate_payload(block_msg.buf, block_msg.unpack(block_msg.buf))
        except ChecksumError:
            logger.debug("Encountered checksum error, which can be caused by duplicate transaction. "
                         "Stop processing block {}", block_hash)
            return
        block_hash = cast(OntObjectHash, block_hash)

        if isinstance(block_msg, BlockOntMessage):
            if block_hash in self._blocks and not waiting_for_recovery:
                self.node.update_current_block_height(
                    block_msg.height(), block_hash
                )
                inv_msg = InvOntMessage(
                    magic=block_msg.magic(),
                    inv_type=InventoryOntType.MSG_BLOCK,
                    blocks=[block_hash],
                )
                self.connection.enqueue_msg(inv_msg)
        # pyre-fixme[25]: `block_msg` has type `OntConsensusMessage`,
        #  assertion `not isinstance(block_msg, bxgateway.messages.ont.consensus_ont_message.OntConsensusMessage)`
        #  will always fail.
        elif isinstance(block_msg, OntConsensusMessage):
            if block_hash in self._blocks and not waiting_for_recovery:
                self.connection.log_info(
                    "Sending consensus message with block hash {} to blockchain node",
                    block_hash,
                )
                self.connection.enqueue_msg(block_msg)
                handling_time, relay_desc = self.node.track_block_from_bdn_handling_ended(block_hash)
                block_stats.add_block_event_by_block_hash(
                    block_hash,
                    BlockStatEventType.BLOCK_SENT_TO_BLOCKCHAIN_NODE,
                    network_num=self.node.network_num,
                    broadcast_type=BroadcastMessageType.CONSENSUS,
                    more_info="{} bytes; Handled in {}; R - {}; {}".format(
                        len(block_msg.rawbytes()),
                        stats_format.duration(handling_time),
                        relay_desc,
                        block_msg.extra_stats_data(),
                    ),
                )

    # pyre-fixme[14]: `bxgateway.services.ont.ont_block_queuing_service.OntBlockQueuingService.send_block_to_node`
    #  overrides method defined in `AbstractBlockQueuingService` inconsistently. Parameter of
    #  type `Optional[Union[BlockOntMessage, OntConsensusMessage]]` is not a supertype of the overridden parameter
    #  `Optional[Variable[bxgateway.services.abstract_block_queuing_service.TBlockMessage
    #  (bound to bxcommon.messages.abstract_block_message.AbstractBlockMessage)]]`.
    def send_block_to_node(
        self,
        block_hash: Sha256Hash,
        block_msg: Optional[Union[BlockOntMessage, OntConsensusMessage]] = None,
    ):
        if block_hash not in self._blocks or block_hash not in self._blocks_waiting_for_recovery:
            return

        waiting_for_recovery = self._blocks_waiting_for_recovery[block_hash]
        if waiting_for_recovery:
            raise Exception(
                "Block {} is waiting for recovery in queuing service for {}".format(block_hash, self.connection.peer_desc)
            )

        block_msg = cast(Union[BlockOntMessage, OntConsensusMessage], self.node.block_storage[block_hash])
        super(OntBlockQueuingService, self).send_block_to_node(
            block_hash, block_msg
        )
        # TODO test remove_from_queue and EncBlockReceivedByGatewayFromNetwork stat event
        self.remove_from_queue(block_hash)

    def update_recovered_block(
        self, block_hash: Sha256Hash, block_msg: Union[BlockOntMessage, OntConsensusMessage]
    ):
        if block_hash not in self._blocks or block_hash not in self._blocks_waiting_for_recovery:
            return

        block_hash = cast(OntObjectHash, block_hash)

        self._blocks_waiting_for_recovery[block_hash] = False
        self.store_block_data(block_hash, block_msg)
        if isinstance(block_msg, BlockOntMessage):
            self.node.update_current_block_height(block_msg.height(), block_hash)
            inv_msg = InvOntMessage(
                magic=block_msg.magic(),
                inv_type=InventoryOntType.MSG_BLOCK,
                blocks=[block_hash],
            )
            self.connection.enqueue_msg(inv_msg)
        # pyre-fixme[25]: `block_msg` has type `OntConsensusMessage`,
        #  assertion `not isinstance(block_msg, bxgateway.messages.ont.consensus_ont_message.OntConsensusMessage)`
        #  will always fail.
        elif isinstance(block_msg, OntConsensusMessage):
            self.connection.enqueue_msg(block_msg)

    def mark_blocks_seen_by_blockchain_node(
        self, block_hashes: List[Sha256Hash]
    ):
        for block_hash in block_hashes:
            self.mark_block_seen_by_blockchain_node(block_hash)

    # pyre-fixme[14]: `bxgateway.services.ont.ont_block_queuing_service.OntBlockQueuingService.mark_block_seen_by_
    #  blockchain_node` overrides method defined in `AbstractBlockQueuingService` inconsistently.
    #  Parameter of type `Optional[Union[BlockOntMessage, OntConsensusMessage]]` is not a supertype of the overridden
    #  parameter `Optional[Variable[bxgateway.services.abstract_block_queuing_service.TBlockMessage
    #  (bound to bxcommon.messages.abstract_block_message.AbstractBlockMessage)]]`.
    def mark_block_seen_by_blockchain_node(
        self,
        block_hash: Sha256Hash,
        block_message: Optional[Union[BlockOntMessage, OntConsensusMessage]] = None,
        _block_number: Optional[int] = None,
    ):
        self._blocks_seen_by_blockchain_node.add(block_hash)

    def _clean_block_queue(self):
        if len(self._block_queue) > ont_constants.ONT_MAX_QUEUED_BLOCKS:
            oldest_block_hash = self._block_queue[0].block_hash
            self.remove(oldest_block_hash)
            self.node.track_block_from_bdn_handling_ended(oldest_block_hash)

    def remove_from_queue(self, block_hash: Sha256Hash) -> int:
        """
        this is similar to the one in abstract_block_queuing_service, with `if block_hash not in self._blocks` removed
        """
        self.connection.log_trace("Removing block {} from queue.", block_hash)

        for index in range(len(self._block_queue)):
            if self._block_queue[index][0] == block_hash:
                del self._block_queue[index]
                del self._blocks_waiting_for_recovery[block_hash]
                return index

        return -1

    def store_block_data(
            self,
            block_hash: Sha256Hash,
            block_msg: Union[BlockOntMessage, OntConsensusMessage]
    ):
        if isinstance(block_msg, BlockOntMessage):
            block_height = block_msg.height()
            if block_height > self._highest_block_number:
                self._highest_block_number = block_height
            self._block_hashes_by_height[block_height] = block_hash
        super().store_block_data(block_hash, block_msg)

    def iterate_block_hashes_starting_from_hash(
            self,
            block_hash: Sha256Hash,
            max_count: int = gateway_constants.TRACKED_BLOCK_MAX_HASH_LOOKUP) -> Iterator[Sha256Hash]:
        """
        iterate over cached blocks headers in descending order
        :param block_hash: starting block hash
        :param max_count: max number of elements to return
        :return: Iterator of block hashes in descending order
        """
        block_hash_ = block_hash
        for _ in range(max_count):
            if block_hash_ and block_hash_ in self._blocks:
                yield block_hash_
                block_msg = self.node.block_storage[block_hash_]
                assert block_msg is not None
                block_hash_ = block_msg.prev_block_hash()
            else:
                break

    def iterate_recent_block_hashes(
            self,
            max_count: int = gateway_constants.TRACKED_BLOCK_MAX_HASH_LOOKUP) -> Iterator[Sha256Hash]:
        """
        :param max_count:
        :return: Iterator[Sha256Hash] in descending order (last -> first)
        """
        if self._highest_block_number not in self._block_hashes_by_height:
            return iter([])
        block_hash = self._block_hashes_by_height[self._highest_block_number]
        return self.iterate_block_hashes_starting_from_hash(block_hash, max_count=max_count)



from enum import auto

from bxcommon.models.serializable_flag import SerializableFlag


# IntFlag allows comparison with ints, which is not as strict as Flag, but allows easier unit testing.
class ConnectionType(SerializableFlag):
    NONE = 0
    SDN = auto()
    BLOCKCHAIN_NODE = auto()
    REMOTE_BLOCKCHAIN_NODE = auto()
    EXTERNAL_GATEWAY = auto()
    RELAY_TRANSACTION = auto()
    RELAY_BLOCK = auto()
    RELAY_ALL = RELAY_TRANSACTION | RELAY_BLOCK
    INTERNAL_GATEWAY = auto()
    GATEWAY = INTERNAL_GATEWAY | EXTERNAL_GATEWAY
    RELAY_PROXY = auto()

    def __str__(self):
        return str(self.name)

    def __format__(self, format_spec):
        return str(self.name)

    def format_short(self):
        cls = self.__class__
        if self in cls.RELAY_ALL:
            return "R"
        if self in cls.BLOCKCHAIN_NODE:
            return "B"
        if self in cls.REMOTE_BLOCKCHAIN_NODE:
            return "RemoteB"
        if self in cls.GATEWAY:
            return "G"

        return self.name



import argparse
import asyncio
import sys
import time
from datetime import datetime
from typing import Optional, Dict, Set, List, Any, cast

from typing import IO

from bloxroute_cli.provider.ws_provider import WsProvider

class HashEntry:
    tx_hash: str
    gateway1_time_received: Optional[float] = None
    gateway2_time_received: Optional[float] = None

    def __init__(
        self, tx_hash: str, gateway1_time_received: Optional[float] = None, gateway2_time_received: Optional[float] = None
    ) -> None:
        self.tx_hash = tx_hash
        self.gateway1_time_received = gateway1_time_received
        self.gateway2_time_received = gateway2_time_received

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, HashEntry)
            and other.tx_hash == self.tx_hash
        )


ws_provider: WsProvider
light_gw_ws_provider: WsProvider
time_to_begin_comparison = 0.0
time_to_end_comparison = 0.0
all_hashes_file: Optional[IO[Any]] = None
missing_hashes_file: Optional[IO[Any]] = None
trail_new_hashes: Set[str] = set()
lead_new_hashes: Set[str] = set()
low_fee_hashes: Set[str] = set()
high_delta_hashes: Set[str] = set()
seen_hashes: Dict[str, HashEntry] = {}
num_intervals_passed = 0
num_intervals: Optional[int] = None


async def main() -> None:
    global seen_hashes
    global trail_new_hashes
    global low_fee_hashes
    global time_to_begin_comparison
    global time_to_end_comparison
    global num_intervals_passed
    global num_intervals

    parser = get_argument_parser()
    args = parser.parse_args()

    if (args.min_gas_price or args.addresses) and args.exclude_tx_contents:
        print("ERROR: If filtering by minimum gas price or addresses, exclude-tx-contents must be False.")
        sys.exit()

    addresses = None if not args.addresses else args.addresses.lower().split(",")
    min_gas = None if not args.min_gas_price else int(args.min_gas_price * 10e8)

    global all_hashes_file
    if "ALL" in args.dump:
        all_hashes_file = open("all_hashes.csv", "w")
        f = all_hashes_file
        assert f is not None
        f.write("tx-hash,blxr1 Time,blxr2 Time\n")

    global missing_hashes_file
    if "MISSING" in args.dump:
        missing_hashes_file = open("missing_hashes.txt", "w")

    time_to_begin_comparison = time.time() + args.lead_time
    time_to_end_comparison = time_to_begin_comparison + args.interval
    num_intervals = args.num_intervals

    asyncio.create_task(process_new_txs_gateway(
        args.gateway1,
        args.auth_header1,
        args.feed_name,
        args.exclude_tx_contents,
        args.exclude_duplicates,
        args.exclude_from_blockchain,
        min_gas,
        addresses,
        args.use_light_gateway1
    ))
    asyncio.create_task(process_new_txs_gateway(
        args.gateway2,
        args.auth_header2,
        args.feed_name,
        args.exclude_tx_contents,
        args.exclude_duplicates,
        args.exclude_from_blockchain,
        min_gas,
        addresses,
        args.use_light_gateway2
    ))

    await asyncio.sleep(args.lead_time)
    while True:
        await asyncio.sleep(args.interval)
        trail_new_hashes.clear()
        await asyncio.sleep(args.trail_time)
        print(
            f"-----------------------------------------------------\n"
            f"Interval: {args.interval} seconds. "
            f"End time: {datetime.now().isoformat(sep=' ', timespec='seconds')}. "
            f"Minimum gas price: {args.min_gas_price} "
            f"{stats(seen_hashes, args.ignore_delta, args.verbose)}"
        )
        seen_hashes.clear()
        lead_new_hashes.clear()
        time_to_end_comparison = time.time() + args.interval

        num_intervals_passed += 1
        if num_intervals_passed == args.num_intervals:
            print(f"{num_intervals_passed} of {num_intervals} intervals complete. Exiting.")
            sys.exit(0)


async def process_new_txs_bdn(
    feed_name: str, exclude_tx_contents: bool, exclude_duplicates: bool, exclude_from_blockchain: bool,
    min_gas: Optional[int], addresses: List[str], use_light_gateway: bool = False
) -> None:
    options: Dict[str, Any] = {}
    if not use_light_gateway:
        options = {
            "duplicates": not exclude_duplicates,
            "include_from_blockchain": not exclude_from_blockchain
        }

    if exclude_tx_contents:
        options["include"] = ["tx_hash"]
    else:
        options["include"] = ["tx_hash", "tx_contents"]
    if use_light_gateway:
        subscription_id = await light_gw_ws_provider.subscribe(feed_name, options)
    else:
        subscription_id = await ws_provider.subscribe(feed_name, options)

    while True:
        if use_light_gateway:
            next_notification = await light_gw_ws_provider.get_next_subscription_notification_by_id(subscription_id)
        else:
            next_notification = await ws_provider.get_next_subscription_notification_by_id(subscription_id)
        time_received = time.time()
        transaction_hash = next_notification.notification["txHash"]

        if time_received < time_to_begin_comparison:
            lead_new_hashes.add(transaction_hash)
            continue

        if not exclude_tx_contents:
            tx_contents = next_notification.notification["txContents"]
            if (
                addresses
                and "to" in next_notification.notification["txContents"]
                and tx_contents["to"] not in addresses
            ):
                continue
            if min_gas and int(tx_contents["gasPrice"], 16) < min_gas:
                low_fee_hashes.add(transaction_hash)
                continue

        if transaction_hash in seen_hashes:
            hash_entry = seen_hashes[transaction_hash]
            if not use_light_gateway:
                if hash_entry.gateway1_time_received is None:
                    hash_entry.gateway1_time_received = time_received
            else:
                if hash_entry.gateway2_time_received is None:
                    hash_entry.gateway2_time_received = time_received
        elif time_received < time_to_end_comparison and transaction_hash not in trail_new_hashes and transaction_hash not in lead_new_hashes:
            if not use_light_gateway:
                seen_hashes[transaction_hash] = HashEntry(transaction_hash, gateway1_time_received=time.time())
            else:
                seen_hashes[transaction_hash] = HashEntry(transaction_hash, gateway2_time_received=time.time())
        else:
            trail_new_hashes.add(transaction_hash)


async def process_new_txs_gateway(
    gateway_url: str, auth_header: str, feed_name: str, exclude_tx_contents: bool, exclude_duplicates: bool, exclude_from_blockchain: bool,
    min_gas: Optional[int], addresses: List[str], use_light_gateway: bool
) -> None:
    print(f"Initiating connection to: {gateway_url}")
    headers = {"Authorization": auth_header}
    async with WsProvider(gateway_url, headers=headers) as ws:
        print(f"websockets endpoint: {gateway_url} established")
        global ws_provider
        global light_gw_ws_provider
        if use_light_gateway:
            light_gw_ws_provider = cast(WsProvider, ws)
        else:
            ws_provider = cast(WsProvider, ws)

        await process_new_txs_bdn(feed_name, exclude_tx_contents, exclude_duplicates, exclude_from_blockchain, min_gas, addresses, use_light_gateway)


def stats(seen_hashes: Dict[str, HashEntry], ignore_delta: int, verbose: bool) -> str:
    tx_seen_by_both_feeds_gateway1_first = 0
    tx_seen_by_both_feeds_gateway2_first = 0
    tx_received_by_gateway1_first_total_delta = 0
    tx_received_by_gateway2_first_total_delta = 0
    new_tx_from_gateway1_feed_first = 0
    new_tx_from_gateway2_feed_first = 0
    total_tx_from_gateway1 = 0
    total_tx_from_gateway2 = 0

    for tx_hash, hash_entry in seen_hashes.items():
        if hash_entry.gateway1_time_received is None:
            gateway2_time_received = hash_entry.gateway2_time_received
            assert isinstance(gateway2_time_received, float)

            if missing_hashes_file:
                f = missing_hashes_file
                assert f is not None
                f.write(f"{tx_hash}\n")
            if all_hashes_file:
                f = all_hashes_file
                assert f is not None
                f.write(f"{tx_hash}, 0, {datetime.fromtimestamp(gateway2_time_received)}\n")
            new_tx_from_gateway2_feed_first += 1
            total_tx_from_gateway2 += 1
            continue
        if hash_entry.gateway2_time_received is None:
            gateway1_time_received = hash_entry.gateway1_time_received
            assert isinstance(gateway1_time_received, float)

            if all_hashes_file:
                f = all_hashes_file
                assert f is not None
                f.write(f"{tx_hash}, {datetime.fromtimestamp(gateway1_time_received)}, 0\n")
            new_tx_from_gateway1_feed_first += 1
            total_tx_from_gateway1 += 1
            continue

        gateway2_time_received = hash_entry.gateway2_time_received
        gateway1_time_received = hash_entry.gateway1_time_received
        assert isinstance(gateway2_time_received, float)
        assert isinstance(gateway1_time_received, float)

        total_tx_from_gateway1 += 1
        total_tx_from_gateway2 += 1

        if abs(gateway1_time_received - gateway2_time_received) > ignore_delta:
            high_delta_hashes.add(hash_entry.tx_hash)
            continue

        if all_hashes_file:
            f = all_hashes_file
            assert f is not None
            f.write(
                f"{tx_hash}, "
                f"{datetime.fromtimestamp(gateway1_time_received)}, "
                f"{datetime.fromtimestamp(gateway2_time_received)}\n"
            )

        if gateway1_time_received < gateway2_time_received:
            new_tx_from_gateway1_feed_first += 1
            tx_seen_by_both_feeds_gateway1_first += 1
            tx_received_by_gateway1_first_total_delta += gateway2_time_received - gateway1_time_received
        elif gateway2_time_received < gateway1_time_received:
            new_tx_from_gateway2_feed_first += 1
            tx_seen_by_both_feeds_gateway2_first += 1
            tx_received_by_gateway2_first_total_delta += gateway1_time_received - gateway2_time_received

    new_tx_seen_by_both_feeds = tx_seen_by_both_feeds_gateway1_first + tx_seen_by_both_feeds_gateway2_first
    tx_received_by_gw1_first_avg_delta = 0 if tx_seen_by_both_feeds_gateway1_first == 0 \
        else round((tx_received_by_gateway1_first_total_delta / float(tx_seen_by_both_feeds_gateway1_first) * 1000), 1)
    tx_received_by_gw2_first_avg_delta = 0 if tx_seen_by_both_feeds_gateway2_first == 0 \
        else round((tx_received_by_gateway2_first_total_delta / float(tx_seen_by_both_feeds_gateway2_first) * 1000), 1)
    percentage_tx_seen_first_by_gateway = 0 if new_tx_seen_by_both_feeds == 0 \
        else int((tx_seen_by_both_feeds_gateway1_first / new_tx_seen_by_both_feeds) * 100)

    results_str = (
        f"\nAnalysis of Transactions received on both feeds:\n"
        f"Number of transactions: {new_tx_seen_by_both_feeds}\n"
        f"Number of transactions received from Gateway1 first: {tx_seen_by_both_feeds_gateway1_first}\n"
        f"Number of transactions received from Gateway2 first: {tx_seen_by_both_feeds_gateway2_first}\n"
        f"Percentage of transactions seen first from Gateway1: {percentage_tx_seen_first_by_gateway}%\n"
        f"Average time difference for transactions received first from Gateway1 (ms): {tx_received_by_gw1_first_avg_delta}\n"
        f"Average time difference for transactions received first from Gateway2 (ms): {tx_received_by_gw2_first_avg_delta}\n"

        f"\nTotal Transactions summary:\n"
        f"Total tx from Gateway1: {total_tx_from_gateway1}\n"
        f"Total tx from Gateway2: {total_tx_from_gateway2}\n"
        f"Number of low fee tx ignored: {len(low_fee_hashes)}\n"
    )

    verbose_results_str = (
        f"Number of high delta tx ignored: {len(high_delta_hashes)}\n"
        f"Number of new transactions received first from Gateway1: {new_tx_from_gateway1_feed_first}\n"
        f"Number of new transactions received first from Gateway2: {new_tx_from_gateway2_feed_first}\n"
        f"Total number of transactions seen: {new_tx_from_gateway2_feed_first + new_tx_from_gateway1_feed_first}\n"
    )

    if verbose:
        results_str = results_str + verbose_results_str
    return results_str


def get_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument("--gateway1", type=str, default="ws://127.0.0.1:28333/ws")
    parser.add_argument("--gateway2", type=str, default="ws://127.0.0.1:28344/ws")
    parser.add_argument("--feed-name", type=str, default="newTxs", choices=["newTxs", "pendingTxs", "transactionStatus"])
    parser.add_argument("--min-gas-price", type=float, default=None, help="Gas price in gigawei")
    parser.add_argument("--addresses", type=str, default=None, help="Comma separated list of Node addresses")
    parser.add_argument("--exclude-tx-contents", action="store_true")
    parser.add_argument("--interval", type=int, default=600, help="Length of feed sample interval in seconds")
    parser.add_argument("--num-intervals", type=int, default=6)
    parser.add_argument("--lead-time", type=int, default=60, help="Seconds to wait before starting to compare feeds")
    parser.add_argument("--trail-time", type=int, default=60, help="Seconds to wait after interval to receive tx on both feeds")
    parser.add_argument("--dump", type=str, default="", choices=["ALL", "MISSING", "ALL,MISSING"])
    parser.add_argument("--exclude-duplicates", action="store_true", default=True, help="For pendingTxs only")
    parser.add_argument("--ignore-delta", type=int, default=5, help="Ignore tx with delta above this amount (seconds)")
    parser.add_argument("--use-cloud-api", action="store_true")
    parser.add_argument("--verbose", action="store_true")
    parser.add_argument("--exclude-from-blockchain", action="store_true")
    parser.add_argument("--cloud-api-ws-uri", type=str, default="wss://api.blxrbdn.com/ws")
    parser.add_argument("--auth-header1", type=str, help="Authorization header created with account id and password for python gateway")
    parser.add_argument("--auth-header2", type=str, help="Authorization header created with account id and password for light gateway")
    parser.add_argument("--use-light-gateway1", action="store_true")
    parser.add_argument("--use-light-gateway2", action="store_true")

    return parser


def run_main():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(main())
    except KeyboardInterrupt:
        print(f"{num_intervals_passed} of {num_intervals} intervals complete. Exiting.")
        sys.exit(0)



class EthProtocolMessageType:
    AUTH = b"auth"
    AUTH_ACK = b"auth_ack"
    HELLO = 0
    DISCONNECT = 1
    PING = 2
    PONG = 3
    STATUS = 16
    NEW_BLOCK_HASHES = 17
    TRANSACTIONS = 18
    GET_BLOCK_HEADERS = 19
    BLOCK_HEADERS = 20
    GET_BLOCK_BODIES = 21
    BLOCK_BODIES = 22
    NEW_BLOCK = 23
    NEW_POOLED_TRANSACTION_HASHES = 24
    GET_POOLED_TRANSACTIONS = 25
    POOLED_TRANSACTIONS = 26
    GET_NODE_DATA = 29
    NODE_DATA = 30
    GET_RECEIPTS = 31
    RECEIPTS = 32

    NEW_BLOCK_BYTES = b"new_block"




from bxgateway.messages.btc.btc_message import BtcMessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType


# xversion message type in Bitcoin Cash
# implemented this type to handle the message processing error
class XversionBtcMessage(BtcMessage):
    MESSAGE_TYPE = BtcMessageType.XVERSION

    def __init__(self, buf: bytearray):
        self.buf = buf
        self._memoryview = memoryview(buf)
        self._magic = self._command = self._payload_len = self._checksum = None
        self._payload = None



from typing import NamedTuple, Union, Dict, Any

from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.feed.new_transaction_feed import FeedSource


class EthRawTransaction(NamedTuple):
    """
    Transaction contents can be any of the following:
     - confirmed transaction from transaction streaming gateway (memoryview)
     - transaction contents from BDN transaction service (memoryview)
     - transaction from Ethereum node socket connection (memoryview)
     - transaction from Ethereum RPC call (JSON)

    Raw pre-serialized transaction must be able to handle all cases.
    """
    tx_hash: Sha256Hash
    tx_contents: Union[memoryview, Dict[str, Any]]
    source: FeedSource
    local_region: bool



from typing import List

import blxr_rlp as rlp
from bxcommon.utils import convert
from bxcommon.utils.object_hash import Sha256Hash
from bxutils.logging.log_level import LogLevel

from bxgateway.messages.eth.protocol.eth_protocol_message import EthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType
from bxcommon.utils.blockchain_utils.eth import rlp_utils


class GetBlockBodiesEthProtocolMessage(EthProtocolMessage):
    msg_type = EthProtocolMessageType.GET_BLOCK_BODIES

    fields = [("block_hashes", rlp.sedes.CountableList(rlp.sedes.binary))]

    def __repr__(self):
        requested_hashes = self.get_field_value("block_hashes")
        request_repr = list(requested_hashes[:1])
        if len(requested_hashes) > 1:
            request_repr.append(requested_hashes[-1])
        request_repr = map(convert.bytes_to_hex, request_repr)
        return (
            f"GetBlockBodiesEthProtocolMessage<"
            f"bodies_count: {len(requested_hashes)}, "
            f"hashes: [{'...'.join([requested_hash for requested_hash in request_repr])}]>"
        )

    def get_block_hashes(self) -> List[Sha256Hash]:
        if self._memory_view is None:
            self.serialize()

        return list(
            map(
                Sha256Hash,
                rlp_utils.get_first_list_field_items_bytes(
                    self._memory_view, remove_items_length_prefix=True
                )
            )
        )

    def log_level(self):
        return LogLevel.DEBUG



import struct

from bxcommon import constants
from bxcommon.messages.abstract_internal_message import AbstractInternalMessage
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.abstract_broadcast_message import AbstractBroadcastMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.broadcast_message import BroadcastMessage
from bxcommon.messages.bloxroute.v8.broadcast_message_v8 import BroadcastMessageV8
from bxcommon.messages.versioning.abstract_message_converter import AbstractMessageConverter
from bxcommon.models.broadcast_message_type import BroadcastMessageType


class _BroadcastMessageConverterV8(AbstractMessageConverter):
    _MSG_TYPE_TO_OLD_MSG_CLASS_MAPPING = {
        BloxrouteMessageType.BROADCAST: BroadcastMessageV8
    }

    _MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING = {
        BloxrouteMessageType.BROADCAST: BroadcastMessage
    }

    _BASE_LENGTH = (
        AbstractBroadcastMessage.HEADER_LENGTH
        + AbstractBroadcastMessage.PAYLOAD_LENGTH
        - constants.CONTROL_FLAGS_LEN
    )

    _BREAKPOINT = (
        _BASE_LENGTH + constants.BROADCAST_TYPE_LEN
    )

    def convert_to_older_version(
        self, msg: AbstractInternalMessage
    ) -> AbstractInternalMessage:
        msg_type = msg.MESSAGE_TYPE

        if msg_type not in self._MSG_TYPE_TO_OLD_MSG_CLASS_MAPPING:
            raise ValueError(
                f"Tried to convert unexpected new message type to v8: {msg_type}"
            )

        old_version_msg_class = self._MSG_TYPE_TO_OLD_MSG_CLASS_MAPPING[
            msg_type
        ]
        old_version_payload_len = msg.payload_len() - constants.BROADCAST_TYPE_LEN

        old_version_msg_bytes = bytearray(self._BASE_LENGTH + len(msg.rawbytes()[self._BREAKPOINT:]))
        old_version_msg_bytes[:self._BASE_LENGTH] = msg.rawbytes()[:self._BASE_LENGTH]
        old_version_msg_bytes[self._BASE_LENGTH:] = msg.rawbytes()[self._BREAKPOINT:]

        return AbstractBloxrouteMessage.initialize_class(
            old_version_msg_class,
            old_version_msg_bytes,
            (msg_type, old_version_payload_len),
        )

    def convert_from_older_version(
        self, msg: AbstractInternalMessage
    ) -> AbstractInternalMessage:
        msg_type = msg.MESSAGE_TYPE

        if msg_type not in self._MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING:
            raise ValueError(
                f"Tried to convert unexpected old message type to v9: {msg_type}"
            )

        new_msg_class = self._MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING[msg_type]
        new_payload_len = msg.payload_len() + constants.BROADCAST_TYPE_LEN

        new_msg_bytes = bytearray(AbstractBloxrouteMessage.HEADER_LENGTH + new_payload_len)
        new_msg_bytes[:self._BASE_LENGTH] = msg.rawbytes()[:self._BASE_LENGTH]
        struct.pack_into(
            "<4s",
            new_msg_bytes,
            self._BASE_LENGTH,
            # pylint: disable=no-member
            BroadcastMessageType.BLOCK.value.encode(constants.DEFAULT_TEXT_ENCODING)
        )
        new_msg_bytes[self._BREAKPOINT:] = msg.rawbytes()[self._BASE_LENGTH:]

        return AbstractBloxrouteMessage.initialize_class(
            new_msg_class,
            new_msg_bytes,
            (msg_type, new_payload_len)
        )

    def convert_first_bytes_to_older_version(
        self, first_msg_bytes: memoryview
    ) -> memoryview:
        if len(first_msg_bytes) < AbstractBloxrouteMessage.HEADER_LENGTH + AbstractBroadcastMessage.PAYLOAD_LENGTH - \
                constants.CONTROL_FLAGS_LEN:
            raise ValueError("Not enough bytes to convert.")
        command, payload_len = BroadcastMessage.unpack(first_msg_bytes)

        result_bytes = bytearray(len(first_msg_bytes) - constants.BROADCAST_TYPE_LEN)

        result_bytes[:self._BASE_LENGTH] = first_msg_bytes[:self._BASE_LENGTH]
        result_bytes[self._BASE_LENGTH:] = first_msg_bytes[self._BREAKPOINT:]

        struct.pack_into("<12sL", result_bytes, constants.STARTING_SEQUENCE_BYTES_LEN, command,
                         payload_len - constants.BROADCAST_TYPE_LEN)

        return memoryview(result_bytes)

    def convert_first_bytes_from_older_version(
        self, first_msg_bytes: memoryview
    ) -> memoryview:
        if len(first_msg_bytes) < AbstractBloxrouteMessage.HEADER_LENGTH + AbstractBroadcastMessage.PAYLOAD_LENGTH - \
                constants.CONTROL_FLAGS_LEN:
            raise ValueError("Not enough bytes to convert.")
        command, payload_len = BroadcastMessageV8.unpack(first_msg_bytes)

        result_bytes = bytearray(len(first_msg_bytes) + constants.BROADCAST_TYPE_LEN)

        result_bytes[:self._BASE_LENGTH] = first_msg_bytes[:self._BASE_LENGTH]
        result_bytes[self._BREAKPOINT:] = first_msg_bytes[self._BASE_LENGTH:]
        struct.pack_into(
            "<4s",
            result_bytes,
            self._BASE_LENGTH,
            # pylint: disable=no-member
            BroadcastMessageType.BLOCK.value.encode(constants.DEFAULT_TEXT_ENCODING)
        )

        struct.pack_into("<12sL", result_bytes, constants.STARTING_SEQUENCE_BYTES_LEN, command,
                         payload_len + constants.BROADCAST_TYPE_LEN)

        return memoryview(result_bytes)

    def convert_last_bytes_to_older_version(
        self, last_msg_bytes: memoryview
    ) -> memoryview:
        return last_msg_bytes

    def convert_last_bytes_from_older_version(
        self, last_msg_bytes: memoryview
    ) -> memoryview:
        return last_msg_bytes

    def get_message_size_change_to_older_version(self) -> int:
        return -constants.BROADCAST_TYPE_LEN

    def get_message_size_change_from_older_version(self) -> int:
        return constants.BROADCAST_TYPE_LEN


broadcast_message_converter_v8 = _BroadcastMessageConverterV8()



from bxcommon.messages.versioning.abstract_version_manager import AbstractVersionManager
from bxgateway.messages.gateway.v1.gateway_message_factory_v1 import gateway_message_factory_v1
from bxgateway.messages.gateway.gateway_message_factory import gateway_message_factory as gateway_message_factory_v2
from bxgateway.messages.gateway.gateway_message_type import GatewayMessageType
from bxgateway.messages.gateway.gateway_message_convertor_factory import gateway_message_converter_factory_v1


class _GatewayVersionManager(AbstractVersionManager):
    CURRENT_PROTOCOL_VERSION = 2
    _PROTOCOL_TO_CONVERTER_FACTORY_MAPPING = {
        1: gateway_message_converter_factory_v1
    }

    _PROTOCOL_TO_FACTORY_MAPPING = {
        1: gateway_message_factory_v1,
        2: gateway_message_factory_v2,
    }

    def __init__(self):
        super(_GatewayVersionManager, self).__init__()
        self.protocol_to_factory_mapping = self._PROTOCOL_TO_FACTORY_MAPPING
        self.protocol_to_converter_factory_mapping = self._PROTOCOL_TO_CONVERTER_FACTORY_MAPPING
        self.version_message_command = GatewayMessageType.HELLO


gateway_version_manager = _GatewayVersionManager()



import struct
from typing import Union, Optional

from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.models.entity_type_model import EntityType
from bxcommon.models.notification_code import NotificationCode, NotificationCodeRange
from bxcommon.models.notification_code_formatting import NotificationFormatting
from bxutils.logging.log_level import LogLevel


class NotificationMessage(AbstractBloxrouteMessage):
    """
    BloXroute message that contains a notification message

    """
    BASE_PAYLOAD_LENGTH = AbstractBloxrouteMessage.HEADER_LENGTH + constants.UL_SHORT_SIZE_IN_BYTES +\
                          constants.CONTROL_FLAGS_LEN
    MESSAGE_TYPE = BloxrouteMessageType.NOTIFICATION

    def __init__(self, notification_code: Optional[Union[NotificationCode, int]] = None, raw: Optional[str] = None,
                 buf: Optional[bytearray] = None):
        if buf is None:
            buffer_len = self.BASE_PAYLOAD_LENGTH + (len(raw) if raw is not None else 0)
            buf = bytearray(buffer_len)

            off = AbstractBloxrouteMessage.HEADER_LENGTH
            struct.pack_into("<H", buf, off, notification_code)
            off += constants.UL_SHORT_SIZE_IN_BYTES
            if raw is not None:
                raw_ = raw.encode()
                buf[off:off + len(raw_)] = raw_
        else:
            buffer_len = len(buf)

        payload_len = buffer_len - AbstractBloxrouteMessage.HEADER_LENGTH

        self.buf: Optional[bytearray] = buf
        self._raw: Optional[str] = None
        self._notification_code: Optional[NotificationCode] = None
        self._memoryview = memoryview(buf)
        super(NotificationMessage, self).__init__(self.MESSAGE_TYPE, payload_len, buf)

    def __repr__(self):
        return "NotificationMsg: {}".format(self.formatted_message())

    def notification_code(self) -> Union[NotificationCode, int]:
        if self._notification_code is None:
            self._unpack()
        notification_code = self._notification_code
        assert notification_code is not None
        return notification_code

    def raw_message(self) -> str:
        if self._raw is None:
            self._unpack()
        raw = self._raw
        assert raw is not None
        return raw

    def formatted_message(self) -> str:
        if self._notification_code is None:
            self._unpack()
        notification_code = self._notification_code
        assert notification_code is not None
        raw = self._raw

        if self._notification_code == NotificationCode.QUOTA_FILL_STATUS:
            assert raw is not None
            args_list = raw.split(",")
            args_list[1] = str(EntityType(int(args_list[1])))
            return NotificationFormatting[self._notification_code].format(*args_list)

        elif self._notification_code == NotificationCode.ACCOUNT_EXPIRED_NOTIFICATION:
            return NotificationFormatting[self._notification_code]

        elif self._notification_code in NotificationFormatting:
            assert raw is not None
            return NotificationFormatting[self._notification_code].format(
                notification_code.value,
                notification_code.name,
                *raw.split(",")
            )
        else:
            return "{}: {}".format(notification_code, raw)

    def level(self) -> LogLevel:
        if self._notification_code is None:
            self._unpack()
        notification_code = self._notification_code
        assert notification_code is not None

        if notification_code < NotificationCodeRange.DEBUG:
            return LogLevel.DEBUG
        elif notification_code < NotificationCodeRange.INFO:
            return LogLevel.INFO
        elif notification_code < NotificationCodeRange.WARNING:
            return LogLevel.WARNING
        else:
            return LogLevel.ERROR

    def log_level(self) -> LogLevel:
        return LogLevel.DEBUG

    def _unpack(self) -> None:
        buf = self.buf
        assert buf is not None

        off = AbstractBloxrouteMessage.HEADER_LENGTH
        notification_code, = struct.unpack_from("<H", buf, off)
        off += constants.UL_SHORT_SIZE_IN_BYTES

        if notification_code in [item.value for item in NotificationCode]:
            self._notification_code = NotificationCode(notification_code)
        else:
            self._notification_code = notification_code

        self._raw = buf[off:AbstractBloxrouteMessage.HEADER_LENGTH +
                             self.payload_len() - constants.CONTROL_FLAGS_LEN].decode()



import struct
from typing import Optional, Union

from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_broadcast_message import (
    AbstractBroadcastMessage,
)
from bxcommon.messages.bloxroute.bloxroute_message_type import (
    BloxrouteMessageType,
)
from bxcommon.models.quota_type_model import QuotaType
from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.utils.object_hash import Sha256Hash


class TxMessageV17(AbstractBroadcastMessage):
    PAYLOAD_LENGTH = (
        AbstractBroadcastMessage.PAYLOAD_LENGTH
        + constants.SID_LEN
        + constants.TRANSACTION_FLAG_LEN
        + constants.UL_INT_SIZE_IN_BYTES
    )
    MESSAGE_TYPE = BloxrouteMessageType.TRANSACTION
    EMPTY_TX_VAL = memoryview(bytes())

    def __init__(
        self,
        message_hash: Optional[Sha256Hash] = None,
        network_num: Optional[int] = None,
        source_id: str = "",
        short_id: int = constants.NULL_TX_SID,
        tx_val: Union[bytearray, bytes, memoryview, None] = None,
        transaction_flag: Optional[TransactionFlag] = None,
        timestamp: Union[int, float] = constants.NULL_TX_TIMESTAMP,
        buf: Optional[Union[bytearray, memoryview]] = None,
    ):
        self._short_id = None
        self._tx_val: Optional[memoryview] = None
        self._transaction_flag = None
        self._timestamp = None

        # override payload length for variable length message
        if tx_val is not None:
            # pylint: disable=invalid-name
            self.PAYLOAD_LENGTH = (
                AbstractBroadcastMessage.PAYLOAD_LENGTH
                + constants.SID_LEN
                + constants.TRANSACTION_FLAG_LEN
                + constants.UL_INT_SIZE_IN_BYTES
                + len(tx_val)
            )
        super().__init__(message_hash, network_num, source_id, buf)

        if buf is None:
            # minus control flag
            off = (
                self.HEADER_LENGTH
                + AbstractBroadcastMessage.PAYLOAD_LENGTH
                - constants.CONTROL_FLAGS_LEN
            )

            struct.pack_into("<L", self.buf, off, short_id)
            off += constants.SID_LEN

            if transaction_flag is None:
                transaction_flag = TransactionFlag.NO_FLAGS
            struct.pack_into("<H", self.buf, off, transaction_flag.value)
            off += constants.TRANSACTION_FLAG_LEN

            struct.pack_into("<L", self.buf, off, int(timestamp))
            off += constants.UL_INT_SIZE_IN_BYTES

            if tx_val is not None:
                self.buf[off:off + len(tx_val)] = tx_val

    def __repr__(self):
        return (
            f"TxMessage<"
            f"tx_hash: {self.tx_hash()}, "
            f"short_id: {self.short_id()}, "
            f"network_num: {self.network_num()}, "
            f"compact: {self.is_compact()}, "
            f"source_id: {self.source_id_as_str()}, "
            f"transaction_flag: {self.transaction_flag()}, "
            f"timestamp: {self.timestamp()}"
            f">"
        )

    def tx_hash(self) -> Sha256Hash:
        return self.message_hash()

    def short_id(self) -> int:
        if self._short_id is None:
            off = (
                self.HEADER_LENGTH
                + AbstractBroadcastMessage.PAYLOAD_LENGTH
                - constants.CONTROL_FLAGS_LEN
            )
            (self._short_id,) = struct.unpack_from("<L", self.buf, off)
        # pyre-fixme[7]: Expected `int` but got `None`.
        return self._short_id

    def has_short_id(self) -> bool:
        return self.short_id() != constants.NULL_TX_SID

    def transaction_flag(self) -> TransactionFlag:
        if self._transaction_flag is None:
            off = (
                self.HEADER_LENGTH
                + AbstractBroadcastMessage.PAYLOAD_LENGTH
                + constants.SID_LEN
                - constants.CONTROL_FLAGS_LEN
            )

            (transaction_flag, ) = struct.unpack_from("<H", self.buf, off)
            self._transaction_flag = TransactionFlag(transaction_flag)
        assert self._transaction_flag is not None
        return TransactionFlag(self._transaction_flag)

    def set_transaction_flag(self, flag: TransactionFlag) -> None:
        off = (
            self.HEADER_LENGTH
            + AbstractBroadcastMessage.PAYLOAD_LENGTH
            + constants.SID_LEN
            - constants.CONTROL_FLAGS_LEN
        )
        self._transaction_flag = flag
        struct.pack_into("<H", self.buf, off, flag.value)

    def timestamp(self) -> int:
        if self._timestamp is None:
            off = (
                self.HEADER_LENGTH
                + AbstractBroadcastMessage.PAYLOAD_LENGTH
                + constants.SID_LEN
                + constants.TRANSACTION_FLAG_LEN
                - constants.CONTROL_FLAGS_LEN
            )
            (self._timestamp,) = struct.unpack_from("<L", self.buf, off)
        # pyre-fixme[7]: Expected `int` but got `None`.
        return self._timestamp

    def tx_val(self) -> memoryview:
        if self._tx_val is None:
            if self.payload_len() == 0:
                self._tx_val = self.EMPTY_TX_VAL
            else:
                off = (
                    self.HEADER_LENGTH
                    + AbstractBroadcastMessage.PAYLOAD_LENGTH
                    + constants.SID_LEN
                    + constants.TRANSACTION_FLAG_LEN
                    + constants.UL_INT_SIZE_IN_BYTES
                    - constants.CONTROL_FLAGS_LEN
                )
                self._tx_val = self._memoryview[
                    off : self.HEADER_LENGTH
                    + self.payload_len()
                    - constants.CONTROL_FLAGS_LEN
                ]

        tx_val = self._tx_val
        assert tx_val is not None
        return tx_val

    def is_compact(self) -> bool:
        return self.tx_val() == self.EMPTY_TX_VAL

    def set_timestamp(self, timestamp: Union[float, int]):
        timestamp = int(timestamp)
        self._timestamp = timestamp
        off = (
            self.HEADER_LENGTH
            + AbstractBroadcastMessage.PAYLOAD_LENGTH
            + constants.SID_LEN
            + constants.TRANSACTION_FLAG_LEN
            - constants.CONTROL_FLAGS_LEN
        )
        struct.pack_into("<L", self.buf, off, timestamp)

    def clear_short_id(self):
        off = (
            self.HEADER_LENGTH
            + AbstractBroadcastMessage.PAYLOAD_LENGTH
            - constants.CONTROL_FLAGS_LEN
        )
        struct.pack_into("<L", self.buf, off, constants.NULL_TX_SID)
        self._short_id = constants.NULL_TX_SID

    def clear_timestamp(self):
        off = (
            self.HEADER_LENGTH
            + AbstractBroadcastMessage.PAYLOAD_LENGTH
            + constants.SID_LEN
            + constants.TRANSACTION_FLAG_LEN
            - constants.CONTROL_FLAGS_LEN
        )
        struct.pack_into("<L", self.buf, off, constants.NULL_TX_TIMESTAMP)
        self._timestamp = constants.NULL_TX_TIMESTAMP

    def clear_protected_fields(self):
        """
        Clears attributes that should be set if coming from a gateway
        connection.
        """
        self.clear_short_id()
        self.clear_timestamp()

    def quota_type(self) -> QuotaType:
        return self.transaction_flag().get_quota_type()



import logging
from logging import LogRecord
from typing import Type

from bxutils.logging.log_level import LogLevel
from bxutils.logging_messages_utils import LogMessage

logger_class: Type[logging.Logger] = logging.getLoggerClass()
log_record_class: Type[LogRecord] = logging.getLogRecordFactory()  # pyre-ignore


# pyre-fixme[11]: Annotation `log_record_class` is not defined as a type.
class CustomLogRecord(log_record_class):

    def getMessage(self):

        msg = str(self.msg)
        if self.args:
            msg = msg.format(*self.args)
        return msg


# pyre-fixme[11]: Annotation `logger_class` is not defined as a type.
class CustomLogger(logger_class):

    def debug(self, msg, *args, **kwargs):
        self.log(logging.DEBUG, msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.log(logging.INFO, msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.log(logging.WARNING, msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.log(logging.ERROR, msg, *args, **kwargs)

    def exception(self, msg, *args, exc_info=True, **kwargs):
        self.log(logging.ERROR, msg, *args, exc_info=exc_info, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self.log(logging.ERROR, msg, *args, **kwargs)

    def log(self, level, msg, *args, **kwargs):
        """
        Copied from logger class with a skipped type check.
        """
        if self.isEnabledFor(level):
            self._log(level, msg, args, **kwargs)

    def _log(self, level, msg, args, exc_info=None, extra=None, stack_info=False):
        if isinstance(msg, LogMessage):
            if extra is None:
                extra = {}
            extra["code"] = msg.code
            extra["category"] = msg.category
            msg = msg.text
        super(CustomLogger, self)._log(level, msg, args, exc_info, extra, stack_info)

    def fatal(self, msg, *args, exc_info=True, **kwargs):
        if self.isEnabledFor(LogLevel.FATAL):
            self.exception(msg, *args, exc_info=exc_info, **kwargs)

    def stats(self, msg, *args, **kwargs):
        if self.isEnabledFor(LogLevel.STATS):
            self._log(LogLevel.STATS, msg, args, kwargs)

    def statistics(self, msg, *args, **kwargs):
        self.stats(msg, *args, **kwargs)

    def trace(self, msg, *args, **kwargs):
        if self.isEnabledFor(LogLevel.TRACE):
            self._log(LogLevel.TRACE, msg, args, kwargs)

    def set_level(self, level):
        self.setLevel(level)

    def set_immediate_flush(self, flush_immediately: bool):
        pass



from typing import Optional, Union
import blxr_rlp as rlp

from bxcommon.utils.blockchain_utils.eth import crypto_utils, rlp_utils
from bxcommon.utils.blockchain_utils.eth.eth_common_utils import keccak_hash
from bxcommon.messages.eth.serializers.transaction import Transaction
from bxcommon.models.tx_validation_status import TxValidationStatus


def verify_eth_transaction_signature(transaction: Transaction) -> bool:
    """
    checks eth transaction signature
    :param transaction:
    :return: if signature matches public key
    """
    try:
        signature = transaction.signature()
        unsigned_msg = transaction.get_unsigned()
        public_key = crypto_utils.recover_public_key(unsigned_msg, signature, keccak_hash)
        return crypto_utils.verify_signature(public_key, signature, keccak_hash(memoryview(unsigned_msg)))
    # pylint: disable=broad-except
    except Exception:
        return False


def normalize_typed_transaction(tx_bytes: memoryview) -> memoryview:
    """
    RLP could be either [transaction_type][encoded transaction], or
    RLP([transaction_type][encoded transaction]). The first form cannot be
    `rlp.decode`d, as it will only return the transaction type or throw an
    exception if strict=True. Force it to be of the second type.
    """
    item_type, item_length, _item_start = rlp_utils.consume_length_prefix(tx_bytes, 0)
    if item_type == str and item_length == 1:
        tx_bytes = memoryview(rlp.encode(tx_bytes.tobytes()))

    return tx_bytes


def parse_transaction(tx_bytes: memoryview) -> Optional[Transaction]:
    """
    :param tx_bytes: transaction bytes
    :return: if transaction successfully parsed returns None else transaction
    """
    try:
        tx_bytes = normalize_typed_transaction(tx_bytes)
        return rlp.decode(tx_bytes.tobytes(), Transaction)
    # pylint: disable=broad-except
    except Exception:
        return None


def validate_transaction(
    tx_bytes: Union[bytearray, memoryview],
    min_tx_network_fee: int
) -> TxValidationStatus:
    """
    check if transaction is validated - signature is correct and format is valid
    :param tx_bytes:
    :param min_tx_network_fee: int
    :return:
    """
    tx_bytes = bytearray(normalize_typed_transaction(memoryview(tx_bytes)))

    try:
        transaction: Transaction = rlp.decode(tx_bytes, Transaction)
        if verify_eth_transaction_signature(transaction):
            if transaction.adjusted_gas_price() >= min_tx_network_fee:
                return TxValidationStatus.VALID_TX
            else:
                return TxValidationStatus.LOW_FEE
        else:
            return TxValidationStatus.INVALID_SIGNATURE
    except rlp.DecodingError:
        return TxValidationStatus.INVALID_FORMAT



import subprocess
from typing import List, NamedTuple

from bxcommon import constants
from bxcommon.models.outbound_peer_model import OutboundPeerModel
from bxcommon.utils.concurrency.thread_pool import ThreadPool
from bxutils import log_messages
from bxutils import logging

logger = logging.get_logger(__name__)


class NodeLatencyInfo(NamedTuple):
    node: OutboundPeerModel
    latency: float


def get_ping_latency(outbound_peer: OutboundPeerModel) -> NodeLatencyInfo:
    """
    returns ping latency to the outbound peer
    :param outbound_peer: peer to ping
    """
    try:
        with subprocess.Popen(["ping", "-c", "1", outbound_peer.ip], stdout=subprocess.PIPE) as proc:
            try:
                if proc:
                    output = proc.communicate(timeout=constants.PING_TIMEOUT_S)[0].decode()
                    ping_latency = float(output.split("time=", 1)[1].split("ms", 1)[0])
                else:
                    ping_latency = constants.PING_TIMEOUT_S * 1000
                    logger.debug("Could not ping {} {}.", outbound_peer.node_type, outbound_peer.ip)
            except subprocess.TimeoutExpired:
                ping_latency = constants.PING_TIMEOUT_S * 1000
                logger.debug("Ping to {} {} timed out.", outbound_peer.node_type, outbound_peer.ip)
    # pylint: disable=broad-except
    except Exception as e:
        logger.error(
            log_messages.PING_TRIGGERED_AN_ERROR, outbound_peer.node_type, outbound_peer.ip, e
        )
        ping_latency = constants.PING_TIMEOUT_S * 1000

    return NodeLatencyInfo(outbound_peer, ping_latency)


def get_ping_latencies(outbound_peers: List[OutboundPeerModel]) -> List[NodeLatencyInfo]:
    with ThreadPool(len(outbound_peers), "ping") as executor:
        futures = [executor.submit(get_ping_latency, ip) for ip in outbound_peers]
        results = [future.result() for future in futures]
    return results



from abc import ABC
from typing import Optional

from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.utils import crypto
from bxcommon.utils.object_hash import Sha256Hash


class BlockHashMessage(AbstractBloxrouteMessage, ABC):
    PAYLOAD_LENGTH = crypto.SHA256_HASH_LEN + constants.CONTROL_FLAGS_LEN

    def __init__(self, block_hash=None, buf=None) -> None:
        if buf is None:
            buf = bytearray(self.HEADER_LENGTH + self.PAYLOAD_LENGTH)

            off = self.HEADER_LENGTH
            buf[off:off + crypto.SHA256_HASH_LEN] = block_hash.binary
            off += crypto.SHA256_HASH_LEN

        self.buf = buf
        self._block_hash: Optional[Sha256Hash] = None
        super(BlockHashMessage, self).__init__(self.MESSAGE_TYPE, self.PAYLOAD_LENGTH, buf)

    def block_hash(self) -> Sha256Hash:
        block_hash = self._block_hash
        if block_hash is None:
            off = self.HEADER_LENGTH
            block_hash = Sha256Hash(self._memoryview[off:off + crypto.SHA256_HASH_LEN])
            self._block_hash = block_hash
        return block_hash



import asyncio
import uuid
import dataclasses
from typing import Generic, TypeVar, Union, Dict, Any, List, Callable

from bxcommon import constants
from bxcommon.feed import filter_parsing

T = TypeVar("T")


class Subscriber(Generic[T]):
    """
    Subscriber object for asynchronous listening to a feed, with functionality
    to filter received messages on certain fields.

    Any object can published on a feed (and received by this subscriber), though
    users should be careful about ensuring that the object is serializable if
    the object should be received over the websocket or IPC subscriptions.

    If `options` is None, the message will be passed onward with no changes.

    Make sure to update documentation page with any input format change
    """

    subscription_id: str
    messages: "asyncio.Queue[Union[T, Dict[str, Any]]]"
    options: Dict[str, Any]
    filters: str
    validator: Callable[[Dict], bool]
    should_exit: bool

    def __init__(self, options: Dict[str, Any]) -> None:
        self.options = options
        self.subscription_id = str(uuid.uuid4())
        self.messages = asyncio.Queue(constants.RPC_SUBSCRIBER_MAX_QUEUE_SIZE)
        filters = options.get("filters", None)
        self.filters = filters if filters else ""
        self.validator = filter_parsing.get_validator(self.filters)
        self.should_exit = False

    async def receive(self) -> Union[T, Dict[str, Any]]:
        """
        Receives the next message in the queue.

        This function will block until a new message is posted to the queue.
        """
        message = await self.messages.get()
        return message

    def queue(self, message: T) -> Union[T, Dict[str, Any]]:
        """
        Queues up a message, releasing all receiving listeners.

        If too many messages are queued without a listener, this task
        will eventually fail and must be handled.
        """
        include_fields = self.options.get("include", None)
        if include_fields:
            if isinstance(message, dict):
                filtered_message = filter_message(message, include_fields)

            elif dataclasses.is_dataclass(message):
                filtered_message = filter_message(dataclasses.asdict(message), include_fields)
            else:
                filtered_message = filter_message(message.__dict__, include_fields)

            self.messages.put_nowait(filtered_message)
            return filtered_message
        else:
            if hasattr(message, "__dict__"):
                self.messages.put_nowait(message.__dict__)
                return message.__dict__
            else:
                self.messages.put_nowait(message)
                return message

    def fast_queue(self, message: Union[T, Dict[str, Any]]) -> None:
        """
        Queue a message, skipping include validation.

        This method should be used judiciously, ensuring the expected content is
        added. The goal is have objects with a common Python ID for later reference
        when serializing.
        """
        self.messages.put_nowait(message)

    def same_options(self, other: "Subscriber") -> bool:
        self_includes = self.options.get("include", None)
        other_includes = other.options.get("include", None)
        return self_includes == other_includes

    def validate(self, state: Dict[str, Any]) -> bool:
        should_publish = False
        try:
            should_publish = self.validator(state)
        except TypeError:
            self.should_exit = True
        return should_publish


def filter_message(message: Dict[str, Any], include_fields: List[str]) -> Dict[str, Any]:
    output = {}
    for key in include_fields:
        value = message
        partial_message = output
        field_keys = key.split(".")
        field = field_keys[0]
        if field not in value:
            continue
        sub_value = value.get(field)
        if len(field_keys) > 1:
            partial_message = partial_message.setdefault(field, {})
            field = field_keys[1]
            # pyre-fixme[16]: Optional type has no attribute `get`
            if field not in sub_value:
                continue
            sub_value = sub_value.get(field)
        partial_message = partial_message.setdefault(field, sub_value)
    return output



import asyncio
from typing import Iterable, AsyncIterable, Union, Optional, Dict, Any, Tuple

import websockets

from bxcommon.rpc.provider.abstract_ws_provider import AbstractWsProvider
from bxcommon.rpc.rpc_errors import RpcError
from bxcommon.rpc.ws.ws_client import WsClient


class MockWebSocket(websockets.WebSocketClientProtocol):
    def __init__(self):
        super().__init__()
        self.alive_event = asyncio.Event()
        self.send_messages = []
        self.recv_messages: 'asyncio.Queue[websockets.Data]' = asyncio.Queue()
        self.transfer_data_exc = None

    async def wait_closed(self) -> None:
        await self.alive_event.wait()

    async def close(self, code: int = 1000, reason: str = "") -> None:
        self.alive_event.set()

    async def send(
        self,
        message: Union[websockets.Data, Iterable[websockets.Data], AsyncIterable[websockets.Data]],
    ) -> None:
        self.send_messages.append(message)

    async def recv(self) -> websockets.Data:
        return await self.recv_messages.get()


class MockWsProvider(AbstractWsProvider):
    def __init__(self, uri: str):
        super().__init__(uri)
        self.fail_connect = False

    async def subscribe(self, channel: str, options: Optional[Dict[str, Any]] = None) -> str:
        pass

    async def unsubscribe(self, subscription_id: str) -> Tuple[bool, Optional[RpcError]]:
        pass

    async def connect_websocket(self) -> websockets.WebSocketClientProtocol:
        if self.fail_connect:
            raise ConnectionRefusedError
        return MockWebSocket()

    # just for autocompletion
    def get_test_websocket(self) -> Optional[websockets.client.WebSocketClientProtocol]:
        return self.ws


class MockWsClient(WsClient):
    def __init__(self, uri: str):
        super().__init__(uri)
        self.fail_connect = False

    async def subscribe(self, channel: str, options: Optional[Dict[str, Any]] = None) -> str:
        pass

    async def unsubscribe(self, subscription_id: str) -> Tuple[bool, Optional[RpcError]]:
        pass

    async def connect_websocket(self) -> websockets.WebSocketClientProtocol:
        self.ws = MockWebSocket()
        self.connected_event.set()
        wait_event = asyncio.Event()
        self.ws_status_check = asyncio.create_task(
            wait_event.wait()
        )
        ws = self.ws
        assert ws is not None
        return ws

    async def close(self) -> None:
        self.running = False
        super().close()

    # just for autocompletion
    def get_test_websocket(self) -> Optional[websockets.client.WebSocketClientProtocol]:
        return self.ws



from abc import ABC, abstractmethod
from typing import Union, Optional, NamedTuple

from bxcommon.utils.object_hash import Sha256Hash


class BlockValidationResult(NamedTuple):
    is_valid: bool
    block_hash: Optional[Sha256Hash]
    reason: Optional[str]


class AbstractBlockValidator(ABC):
    @abstractmethod
    def validate_block_header(
        self,
        block_header_bytes: Union[bytearray, memoryview],
        last_confirmed_block_number: Optional[int],
        last_confirmed_block_difficulty: Optional[int]
    ) -> BlockValidationResult:
        pass



import binascii
from binascii import Error
from typing import Dict, Optional, Union

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.models.node_type import NodeType

NODE_TO_CONNECTION_TYPE: Dict[NodeType, Dict[NodeType, ConnectionType]] = {
    NodeType.GATEWAY: {
        NodeType.GATEWAY: ConnectionType.GATEWAY,
        NodeType.INTERNAL_GATEWAY: ConnectionType.GATEWAY,
        NodeType.EXTERNAL_GATEWAY: ConnectionType.GATEWAY,
        NodeType.RELAY_TRANSACTION: ConnectionType.RELAY_ALL,
        NodeType.RELAY_BLOCK: ConnectionType.RELAY_ALL,
        NodeType.RELAY: ConnectionType.RELAY_ALL
    },
    NodeType.INTERNAL_GATEWAY: {
        NodeType.GATEWAY: ConnectionType.GATEWAY,
        NodeType.INTERNAL_GATEWAY: ConnectionType.INTERNAL_GATEWAY,
        NodeType.EXTERNAL_GATEWAY: ConnectionType.EXTERNAL_GATEWAY,
        NodeType.RELAY_TRANSACTION: ConnectionType.RELAY_ALL,
        NodeType.RELAY_BLOCK: ConnectionType.RELAY_ALL,
        NodeType.RELAY: ConnectionType.RELAY_ALL,
        NodeType.RELAY_PROXY: ConnectionType.RELAY_ALL,
    },
    NodeType.EXTERNAL_GATEWAY: {
        NodeType.GATEWAY: ConnectionType.GATEWAY,
        NodeType.INTERNAL_GATEWAY: ConnectionType.INTERNAL_GATEWAY,
        NodeType.EXTERNAL_GATEWAY: ConnectionType.EXTERNAL_GATEWAY,
        NodeType.RELAY_TRANSACTION: ConnectionType.RELAY_ALL,
        NodeType.RELAY_BLOCK: ConnectionType.RELAY_ALL,
        NodeType.RELAY: ConnectionType.RELAY_ALL,
        NodeType.RELAY_PROXY: ConnectionType.RELAY_ALL,
    },
    NodeType.RELAY_BLOCK: {
        NodeType.GATEWAY: ConnectionType.GATEWAY,
        NodeType.INTERNAL_GATEWAY: ConnectionType.INTERNAL_GATEWAY,
        NodeType.EXTERNAL_GATEWAY: ConnectionType.EXTERNAL_GATEWAY,
        NodeType.RELAY_BLOCK: ConnectionType.RELAY_BLOCK,
        NodeType.RELAY_PROXY: ConnectionType.RELAY_PROXY,
        NodeType.RELAY: ConnectionType.RELAY_ALL,
        NodeType.API_SOCKET: ConnectionType.SDN
    },
    NodeType.RELAY_TRANSACTION: {
        NodeType.GATEWAY: ConnectionType.GATEWAY,
        NodeType.INTERNAL_GATEWAY: ConnectionType.INTERNAL_GATEWAY,
        NodeType.EXTERNAL_GATEWAY: ConnectionType.EXTERNAL_GATEWAY,
        NodeType.RELAY_TRANSACTION: ConnectionType.RELAY_TRANSACTION,
        NodeType.RELAY_PROXY: ConnectionType.RELAY_PROXY,
        NodeType.RELAY: ConnectionType.RELAY_ALL,
        NodeType.API_SOCKET: ConnectionType.SDN
    },
    NodeType.RELAY: {
        NodeType.GATEWAY: ConnectionType.GATEWAY,
        NodeType.INTERNAL_GATEWAY: ConnectionType.INTERNAL_GATEWAY,
        NodeType.EXTERNAL_GATEWAY: ConnectionType.EXTERNAL_GATEWAY,
        NodeType.RELAY: ConnectionType.RELAY_ALL,
        NodeType.RELAY_PROXY: ConnectionType.RELAY_PROXY,
        NodeType.API_SOCKET: ConnectionType.SDN,
    },
}


def str_to_bool(value: str, default: bool = False) -> bool:
    if default:
        return value not in ["False", "false", "0"]
    else:
        return value in ["True", "true", "1"]


def bytes_to_hex_string_format(bytes_to_convert: Union[bytearray, bytes]) -> str:
    return f"0x{bytes_to_hex(bytes_to_convert)}"


def bytes_to_hex(s: Union[bytes, bytearray, memoryview]) -> str:
    """
    Encodes bytes to hex
    :param s: bytes
    :return: HEX representation of bytes
    """
    # pyre-ignore[6]
    return binascii.hexlify(s).decode("utf-8")


def hex_to_bytes(s: Union[str, bytes, bytearray, memoryview]) -> bytes:
    """
    Decodes hex string to bytes
    :param s: hex string
    :return: bytes
    """

    try:
        # pyre-ignore[6]
        return binascii.unhexlify(s)
    except Error as e:
        raise ValueError(f"Invalid hex string provided: {s}.") from e


def peer_node_to_connection_type(node_type: NodeType, peer_node_type: Optional[NodeType]) -> ConnectionType:
    if peer_node_type is None:
        raise ValueError("peer node type can't be None!")
    return NODE_TO_CONNECTION_TYPE[node_type][peer_node_type]



from collections import deque
from typing import TypeVar, Generic, Deque, Set

T = TypeVar("T")


class LimitedSizeSet(Generic[T]):
    """
    Set with expiration time.
    """

    contents: Set[T]
    _item_tracker: Deque[T]
    _max_size: int

    def __init__(self, max_size: int) -> None:
        self.contents = set()
        self._item_tracker = deque()
        self._max_size = max_size

    def __contains__(self, item: T) -> bool:
        return item in self.contents

    def __iter__(self):
        for item in self.contents:
            yield item

    def remove(self, value: T) -> None:
        """
        Deletion is fairly expensive. Generally, there is no need to delete
        from this data structure.
        """
        self.contents.remove(value)
        self._item_tracker.remove(value)

    def __len__(self) -> int:
        return len(self.contents)

    def add(self, value: T) -> None:
        self.contents.add(value)
        if len(self.contents) != len(self._item_tracker):
            self._make_room_for_new_item()
            self._item_tracker.append(value)

    def _make_room_for_new_item(self) -> None:
        while len(self.contents) > self._max_size:
            key = self._item_tracker.popleft()
            self.contents.remove(key)



from bxcommon.messages.versioning.abstract_version_converter_factory import AbstractMessageConverterFactory
from bxcommon.messages.versioning.no_changes_message_converter import no_changes_message_converter


class _MessageConverterFactoryV22(AbstractMessageConverterFactory):
    _MESSAGE_CONVERTER_MAPPING = {}

    def get_message_converter(self, msg_type):
        if not msg_type:
            raise ValueError("msg_type is required.")

        if msg_type not in self._MESSAGE_CONVERTER_MAPPING:
            return no_changes_message_converter

        return self._MESSAGE_CONVERTER_MAPPING[msg_type]


message_converter_factory_v22 = _MessageConverterFactoryV22()



import time

from collections import OrderedDict
from typing import TypeVar, Generic, Optional, Callable, Dict, Any

T = TypeVar("T")


class ExpirationQueue(Generic[T]):
    """
    Handles queue of item that need to be expired and removed from the queue over time
    """

    time_to_live_sec: int
    # NOTE: this cannot be annotated as an collections.OrderedDict
    queue: Dict[T, float]

    def __init__(self, time_to_live_sec: int) -> None:

        if time_to_live_sec < 0:
            raise ValueError("Time to live cannot be negative.")

        self.time_to_live_sec = time_to_live_sec
        self.queue = OrderedDict()

    def __len__(self) -> int:
        return len(self.queue)

    def __bool__(self) -> bool:
        return len(self) > 0

    def add(self, item: T) -> None:
        """
        Adds item to the queue
        :param item: item
        """
        self.queue[item] = time.time()

    def remove(self, item: T) -> None:
        """
        Removes item from expiration queue
        :param item: item to remove
        """
        if item in self.queue:
            del self.queue[item]

    def remove_expired(
        self,
        current_time: Optional[float] = None,
        remove_callback: Optional[Callable[[T], Any]] = None,
        limit: Optional[int] = None,
        **kwargs
    ):
        """
        Removes expired items from the queue
        :param current_time: time to use as current time for expiration
        :param remove_callback: reference to a callback function that is being called when item is removed
        :param limit: max number of entries to remove in one call
        :param kwargs: keyword args to pass into the callback method
        """
        if current_time is None:
            current_time = time.time()
        if limit is None:
            limit = len(self.queue)

        iterations = 0
        while (
            len(self.queue) > 0 and
            iterations < limit and
            # pyre-fixme[6]: Expected `float` for 1st param but got `Optional[float]`.
            current_time - self.get_oldest_item_timestamp() > self.time_to_live_sec
        ):
            # noinspection PyArgumentList
            # pyre-fixme[28]: Unexpected keyword argument `last`.
            item, _timestamp = self.queue.popitem(last=False)

            if remove_callback is not None:
                remove_callback(item, **kwargs)

            iterations += 1

    def get_oldest(self) -> Optional[T]:
        """
        Returns the value of oldest item in the queue
        :return: value of oldest item
        """
        if not self.queue:
            return None

        return next(iter(self.queue.keys()))

    def get_oldest_item_timestamp(self) -> Optional[float]:
        """
        Returns timestamp of the oldest item
        :return: timestamp of the oldest item
        """
        if not self.queue:
            return None

        oldest_item = self.get_oldest()
        assert oldest_item is not None
        return self.queue[oldest_item]

    def remove_oldest(
        self, remove_callback: Optional[Callable[[T], None]] = None, **kwargs
    ) -> None:
        """
        Remove one oldest item from the queue
        :param remove_callback: reference to a callback function that is being called when item is removed
        """
        if len(self.queue) > 0:
            # pyre-fixme[28]: Unexpected keyword argument `last`.
            item, _timestamp = self.queue.popitem(last=False)

            if remove_callback is not None:
                remove_callback(item, **kwargs)

    def clear(self) -> None:
        self.queue.clear()



from cryptography.x509 import ObjectIdentifier


class ExtensionsObjectIds:
    NODE_TYPE = ObjectIdentifier("1.22.333.4444")
    NODE_ID = ObjectIdentifier("1.22.333.4445")
    ACCOUNT_ID = ObjectIdentifier("1.22.333.4446")
    NODE_PRIVILEGES = ObjectIdentifier("1.22.333.4447")



import time
from collections import deque
from typing import Deque, Optional

from bxcommon.messages.abstract_message import AbstractMessage


class BlockchainMessageQueue:
    """
    Message queue with defined TTL from first object insertion.

    The intention of this queue is to forward queued up messages only if the queue is emptied within the TTL.
    (e.g. if the blockchain connection doesn't recovery fast enough the messages are dropped)
    """
    _queue: Deque[AbstractMessage]
    _time_to_live: int
    _first_item_insertion_time: Optional[float]

    def __init__(self, time_to_live: int):
        self._time_to_live = time_to_live
        self._initialize_queue()

    def append(self, message: AbstractMessage):
        if self._first_item_insertion_time is None:
            self._first_item_insertion_time = time.time()

        if not self._has_queue_expired():
            self._queue.append(message)

    def pop_items(self) -> Deque[AbstractMessage]:
        self._clear_if_needed()
        return_value = self._queue
        self._initialize_queue()
        return return_value

    def _initialize_queue(self):
        self._queue = deque()
        self._first_item_insertion_time = None

    def _clear_if_needed(self):
        if self._has_queue_expired():
            self._queue.clear()

    def _has_queue_expired(self) -> bool:
        first_item_insertion_time = self._first_item_insertion_time
        return (
            first_item_insertion_time is not None and
            time.time() - first_item_insertion_time > self._time_to_live
        )



from bxgateway.messages.gateway.gateway_message_type import GatewayMessageType
from bxgateway.messages.gateway.v1.gateway_hello_message_converter_v1 import gateway_hello_message_converter_v1
from bxcommon.messages.versioning.abstract_version_converter_factory import AbstractMessageConverterFactory


class _GatewayMessageConverterFactoryV1(AbstractMessageConverterFactory):

    _MESSAGE_CONVERTER_MAPPING = {
        GatewayMessageType.HELLO: gateway_hello_message_converter_v1,
    }

    def get_message_converter(self, msg_type):
        if not msg_type:
            raise ValueError("msg_type is required.")

        if msg_type not in self._MESSAGE_CONVERTER_MAPPING:
            raise ValueError("Converter for message type '{}' is not defined.".format(msg_type))

        return self._MESSAGE_CONVERTER_MAPPING[msg_type]


gateway_message_converter_factory_v1 = _GatewayMessageConverterFactoryV1()



import websockets

from bloxroute_cli.provider.ws_provider import WsProvider
from bxcommon.utils import config


class IpcProvider(WsProvider):
    """
    Provider that connects to bxgateway's websocket RPC endpoint.
    """

    def __init__(self, ipc_file: str):
        super().__init__(ipc_file)
        self.ipc_path: str = config.get_data_file(ipc_file)

    async def connect_websocket(self) -> websockets.WebSocketClientProtocol:
        return await websockets.unix_connect(self.ipc_path)



from collections import defaultdict
from datetime import datetime


class ClassMemStats:
    def __init__(self) -> None:
        self.timestamp: datetime = datetime.utcnow()
        self.networks = defaultdict(ClassMemNetworks)


class ClassMemNetworks:
    def __init__(self) -> None:
        self.analyzed_objects = defaultdict(ClassMemObjects)


class ClassMemObjects:
    def __init__(self) -> None:
        self.object_item_count = 1
        self.object_size = 0
        self.object_flat_size = 0
        self.object_type = None
        self.size_type = None



import dataclasses

from dataclasses import dataclass
from typing import List

from bxcommon.models.bdn_service_type import BdnServiceType
from bxcommon.models.time_interval_type import TimeIntervalType


@dataclass
class BdnServiceModelBase:
    interval: TimeIntervalType = TimeIntervalType.DAILY
    service_type: BdnServiceType = BdnServiceType.MSG_QUOTA
    limit: int = 0


@dataclass
class FeedServiceModelBase:
    allow_filtering: bool = True
    available_fields: List[str] = dataclasses.field(default_factory=list)



import socket
import struct

from bxcommon import constants
from bxcommon.constants import UL_INT_SIZE_IN_BYTES
from bxcommon.utils.blockchain_utils.btc import btc_common_utils
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import BtcObjectHash
from bxgateway import btc_constants
from bxgateway.btc_constants import BTC_HDR_COMMON_OFF, BTC_SHA_HASH_LEN


def ipaddrport_to_btcbytearray(ip_addr, ip_port):
    try:
        buf = bytearray(btc_constants.BTC_IP_ADDR_PORT_SIZE)
        buf[10] = 0xff
        buf[11] = 0xff
        buf[12:16] = socket.inet_pton(socket.AF_INET, ip_addr)
        buf[16] = ((ip_port >> 8) & 0xFF)
        buf[17] = (ip_port & 0xFF)
    except socket.error:
        try:
            buf = bytearray(btc_constants.BTC_IP_ADDR_PORT_SIZE)
            buf[0:16] = socket.inet_pton(socket.AF_INET6, ip_addr)
            buf[16] = ((ip_port >> 8) & 0xFF)
            buf[17] = (ip_port & 0xFF)
        except socket.error:
            return None
    return buf


# broken
def btcbytearray_to_ipaddrport(btcbytearray):
    if btcbytearray[0:12] == constants.IP_V4_PREFIX[0:12]:
        return socket.inet_ntop(socket.AF_INET, btcbytearray[12:16]), (btcbytearray[16] << 8) | btcbytearray[17]
    else:
        return socket.inet_ntop(socket.AF_INET6, btcbytearray[:16]), (btcbytearray[16] << 8) | btcbytearray[17]


# pack the value into a varint, return how many bytes it took
def pack_int_to_btc_varint(val, buf, off):
    if val < 253:
        struct.pack_into("B", buf, off, val)
        return 1
    elif val <= 65535:
        struct.pack_into("<BH", buf, off, btc_constants.BTC_VARINT_SHORT_INDICATOR, val)
        return 3
    elif val <= 4294967295:
        struct.pack_into("<BI", buf, off, btc_constants.BTC_VARINT_INT_INDICATOR, val)
        return 5
    else:
        struct.pack_into("<BQ", buf, off, btc_constants.BTC_VARINT_LONG_INDICATOR, val)
        return 9


def get_sizeof_btc_varint(val):
    if val < 253:
        return 1
    elif val <= 65535:
        return 3
    elif val <= 4294967295:
        return 5
    else:
        return 9


def get_next_tx_size(buf, off, tail=-1):
    segwit_flag = btc_common_utils.is_segwit(buf, off)

    if segwit_flag:
        return get_next_segwit_tx_size(buf, off, tail)
    else:
        return get_next_non_segwit_tx_size(buf, off, tail)


def get_next_non_segwit_tx_size(buf, off, tail=-1):
    """
    Returns the size of the next transaction in the.
    Returns -1 if there's a parsing error or the end goes beyond the tail.
    """
    end = off + btc_constants.TX_VERSION_LEN

    io_size, _, _ = btc_common_utils.get_tx_io_count_and_size(buf, end, tail)

    if io_size < 0:
        return -1

    end += io_size + btc_constants.TX_LOCK_TIME_LEN

    if end > tail > 0:
        return -1

    return end - off


def get_next_segwit_tx_size(buf, off, tail=-1):
    """
    Returns the size of the next transaction in the.
    Returns -1 if there's a parsing error or the end goes beyond the tail.
    """
    end = off + btc_constants.TX_VERSION_LEN + btc_constants.TX_SEGWIT_FLAG_LEN

    io_size, txin_c, _ = btc_common_utils.get_tx_io_count_and_size(buf, end, tail)

    if io_size < 0:
        return -1

    end += io_size

    for _ in range(txin_c):
        witness_count, size = btc_common_utils.btc_varint_to_int(buf, end)
        end += size

        for _ in range(witness_count):
            witness_len, size = btc_common_utils.btc_varint_to_int(buf, end)
            end += size + witness_len

    end += btc_constants.TX_LOCK_TIME_LEN

    if end > tail > 0:
        return -1

    return end - off


def pack_block_header(buffer: bytearray, version: int, prev_block: BtcObjectHash, merkle_root: BtcObjectHash,
                      timestamp: int, bits: int, block_nonce: int) -> int:
    """
    Packs Bitcoin block header into beginning of buffer
    :param buffer: buffer
    :param version: version
    :param prev_block: previous block hash
    :param merkle_root: merkle tree root hash
    :param timestamp: timestamp
    :param bits: bits
    :param block_nonce: block nonce
    :return: length of header
    """

    off = BTC_HDR_COMMON_OFF
    struct.pack_into("<I", buffer, off, version)
    off += UL_INT_SIZE_IN_BYTES
    buffer[off:off + BTC_SHA_HASH_LEN] = prev_block.get_little_endian()
    off += BTC_SHA_HASH_LEN
    buffer[off:off + BTC_SHA_HASH_LEN] = merkle_root.get_little_endian()
    off += BTC_SHA_HASH_LEN
    struct.pack_into("<III", buffer, off, timestamp, bits, block_nonce)
    off += 3 * UL_INT_SIZE_IN_BYTES

    return off




import socket
from typing import Union, Optional

from bxcommon.utils.blockchain_utils.ont.ont_common_utils import get_txid, ont_varint_to_int
from bxgateway import ont_constants
from bxgateway.messages.btc.btc_messages_util import pack_int_to_btc_varint


def ipport_to_ontbytearray(ip_port: int) -> Optional[bytearray]:
    buf = bytearray(ont_constants.ONT_IP_PORT_SIZE)
    buf[0] = ((ip_port >> 8) & 0xFF)
    buf[1] = (ip_port & 0xFF)
    return buf


def ipaddr_to_ontbytearray(ip_addr: str) -> Optional[bytearray]:
    try:
        buf = bytearray(ont_constants.ONT_IP_ADDR_SIZE)
        buf[10] = 0xff
        buf[11] = 0xff
        buf[12:16] = socket.inet_pton(socket.AF_INET, ip_addr)
    except socket.error:
        try:
            buf = bytearray(ont_constants.ONT_IP_ADDR_SIZE)
            buf[0:16] = socket.inet_pton(socket.AF_INET6, ip_addr)
        except socket.error:
            return None
    return buf


def pack_int_to_ont_varint(val: int, buf: Union[memoryview, bytearray], off: int) -> int:
    return pack_int_to_btc_varint(val, buf, off)


def get_next_tx_size(buffer: Union[memoryview, bytearray], offset: int) -> int:
    if not isinstance(buffer, memoryview):
        buffer = memoryview(buffer)
    _, txid_len = get_txid(buffer[offset:])
    off = txid_len + offset
    sig_length, size = ont_varint_to_int(buffer, off)
    off += size
    for _ in range(sig_length):
        invoke_length, size = ont_varint_to_int(buffer, off)
        off += size + invoke_length
        verify_length, size = ont_varint_to_int(buffer, off)
        off += size + verify_length
    return off - offset
